implementation module CTBase

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Basic texture generators, combinators and transformers

import CTVector, CTFunalg, CTParams


// Translation, scaling are more general and they are in CTFunalg,
//  together with step, ramp, etc. And a generic transf.

rotated_cs :: !a !a v:(.(Vec2 !a)->.b) -> u:(.(Vec2 !a)->.b) | *,+,- a, [u<=v]
rotated_cs co si f = \(V2 x y) -> f (V2 (co*x+si*y) (co*y-si*x))
rotated :: !a -> .(v:(.(Vec2 !a)->.b) -> u:(.(Vec2 !a)->.b)) | *,+,-,cos,sin a, [u<=v]
rotated a = rotated_cs (cos a) (sin a)
transposed :: v:((Vec2 !a)->.b) -> u:((Vec2 !a)->.b), [u<=v]
transposed f = \(V2 x y) -> f (V2 y x)
rotated90 :: v:(.(Vec2 !a)->.b) -> u:(.(Vec2 !a)->.b) | ~ a, [u<=v]
rotated90 f = \(V2 x y) -> f (V2 y (~x))
rotated45 :: v:(.(Vec2 !Real) -> .a) -> u:((Vec2 !Real)->.a), [u<=v]
rotated45 f = \(V2 x y) -> f (V2 (SqrtH*x+SqrtH*y) (SqrtH*y-SqrtH*x))
rotated60 :: v:(.(Vec2 Real) -> .a) -> u:((Vec2 .Real) -> .a), [u<=v]
rotated60 f = \(V2 x y) -> f (V2 (0.5*x+0.866025403784438647*y) 
                                 (0.5*y-0.866025403784438647*x))
reduced :: v:(.(Vec2 !a)->.b) -> u:(.(Vec2 !a)->.b) | -,fromInt,toReal a, [u<=v]
reduced f = \(V2 x y) -> f (V2 (gfrac x) (gfrac y))
redscaled :: .(Vec2 !a) v:(.(Vec2 !a)->.b)->u:(.(Vec2 !a)->.b) | -,/,fromInt,toReal a, [u<=v]
redscaled (V2 sx sy) f = \(V2 x y) -> f (V2 (gfrac (x/sx)) (gfrac (y/sy)))
//   redscaled v = scaled v o reduced
symreduced :: .(!(Vec2 a)-> .b) !.(Vec2 a) -> .b | *,+,-,fromInt,half,one,toReal,two a
symreduced f (V2 x y) = f (V2 (two*gfrac (x+half) - one) 
                              (two*gfrac (y+half) - one))
repliq :: !.Real -> .(v:(!(Vec2 a) -> .b) -> u:(!.(Vec2 a) -> .b)) 
          | *,+,-,/,fromInt,fromReal,half,one,toReal,two a,[u<=v]
repliq s = scaled (fR s) o symreduced

rotatedAbout :: !.(Vec2 !a) !a -> .(v:((Vec2 a) -> .b) -> u:(.(Vec2 a) -> .b)) | *,+,-,cos,sin,~ a, [u<=v]
rotatedAbout p0 ang  // of_a_texture
 = translated p0 o rotated_cs (cos ang) (sin ang) o translated (~p0)

step :: !a -> a | < ,zero,half,one a
step x | x<zero = zero
       | x>zero = one
       = half
pulse :: !a -> a | <,one,zero,half a
pulse x | x<zero || x>one = zero
        | x>zero && x<one = one
        = half
xstep :: !a !a -> !a | <,half,one,zero a
xstep a x | x<a = zero
          | x>a = one
          = half
xpulse :: !a !a !a -> !a | <,half,one,zero a          
xpulse a b x | x<a || x>b = zero
             | x>a && x<b = one
             = half        

trainpulse :: !a -> !a | -,<,fromInt,half,one,toReal, zero a
trainpulse x = xpulse zero half (gfrac x)
symtrain :: (!a -> !b) !a -> !b | *,+,-,fromInt,half,one,toReal,two a
symtrain f x = f (two*gfrac (x+half) - one)
symteeth :: (!a -> !a) | *,+,-,abs,fromInt,half,one,toReal,two a
symteeth = symtrain abs

// More primitive generators and auxiliary functions

filteredabs :: !a !a -> a | *,+,/,<,half, ~ a
filteredabs dxh x 
  = if (dxh<x || x< ~dxh) x (half*(dxh+x*x/dxh))
filteredpulse :: !a !a !a !a -> a | *,+,-,/,<,two,zero a
filteredpulse ed0 ed1 dxh x
  = max zero ((min (x+dxh) ed1 - max (x-dxh) ed0)/(two*dxh))
smoothstep :: !a !a -> .(a -> a) | *,-,/,<,one,three,two, zero a
smoothstep a b = fn where
  dn = one/(b-a)
  fn x
   | x<a = zero
   | x>b = one
   # t=(x-a)*dn
   = t*t*(three-two*t)
hypulse :: !a !a !a -> .(a -> a) | *,+,-,/,<,one,zero a
hypulse a b dx = f where
 dx2 = sqr dx
 a0 = a-dx
 b0 = b+dx
 f x | x<=a0 || x>b0 = zero
     | x<=a = sqr (x-a0) / dx2
     | x<=b = one
     = sqr (b0-x) / dx2

unitdisk :: (a !b) -> !b | *,+,-,<,half,one,zero b & norm2 a
unitdisk p = step (one - norm2 p)
disk :: (a b) b (a b) -> b | <->,norm2 a & *,+,-,<,half,one,zero b
disk cntr r2 p = step (r2 - norm2 (p<->cntr))


// Stripes (horiz. and vert.)
xstripe :: v:(!a -> .b) -> u:((Vec2 !a) -> .b), [u<=v]
xstripe f = \(V2 x _) -> f x
ystripe :: v:(!a -> .b) -> u:((Vec2 !a) -> .b), [u<=v]
ystripe f = \(V2 _ y) -> f y
checkers :: !(Vec2 !a) -> !a | -,<,fromInt,half,one,toReal,zero a
checkers (V2 x y) = if (gfrac x<half) (xpulse zero half (gfrac y)) 
                                  (xpulse half one (gfrac y))
chboard :: !(Vec2 !.Real) -> Real
chboard (V2 x y)
 # n=entier(2.0*y)
 # m=entier(2.0*x)
 = toReal (abs (n pmod 2 - m pmod 2))

// Reg. polygon. Use with step. 
polygon :: !.Int !.(Vec2 a) -> a | *,+,-,/,<,==,atan,cos,fromReal,sin,toReal,zero a
polygon k (V2 x y)
 # kd = fromReal (Dpi/fromInt k)
   kdj= kd * fromReal (floor (toReal (atan2 y x / kd)))
   ks = kdj+kd
   sj = sin kdj
   cj = cos kdj
 = ((x-cj)*(sin ks-sj) - (cos ks-cj)*(y-sj))

// Reg. star. Step included. Should it?
regstar :: !.Int !.Real !.Real -> !.(!(Vec2 !.Real) -> Real)
regstar n r0 r1 = rstar
where
 r02 =: sqr r0
 r12 =: sqr r1
 alpha =: Dpi/fromInt n
 alph2 =: alpha/two
 hax =: rot2 alph2 X2axis
 p0 =: r0*>hax
 nrm =: rot90 (normalize (p0 - r1*>X2axis))
 hplane p = step (nrm<.>(p-p0))
 reflected v = hplane ((two*v<.>hax)*>hax - v)
 rstar p
  # r2 = norm2 p
  | r2>r12 = zero
  | r2<r02 = one
  # (turns,ared) = rsplit (angle p) alpha
  | ared <= alph2 = rotated turns hplane p
  = rotated turns reflected p
triangle :: !.(Vec2 a) !.(Vec2 a) !.(Vec2 a) -> .(!.(Vec2 !a) -> b) 
            | one,zero b & *,+,-,/,<,one,zero a
triangle p0 p1 p2 = f
where 
  sax =: p1-p0
  tax =: p2-p0
  d   = skew sax tax
  f p
    # p=p-p0
    # s=skew p tax / d
    # t=skew sax p / d
    | s>=zero && t>= zero && s+t < one = one
    = zero


// General gradient map, usually for colours
colmap :: [!.(a !b, !b)] .IMethod !b -> a !b | *>,<+> a & *,+,-,/,<,half,one,three,two,fromReal b  
colmap [(i0,a0):r] meth t = mkc i0 a0 r where
  mkc i0 a0 [(i1,a1):q]
    | t < a0 = i0
    | t <= a1 = interpol i0 i1 meth ((t-a0)/(a1-a0))
    = mkc i1 a1 q
  mkc i _ [] = i

// 3D
rotated3_cs :: !a !a !.(Vec3 a) v:(!(Vec3 a)-> .b)-> u:(!.(Vec3 a)-> .b) 
   | *,+,-,~a,[u<=v]
rotated3_cs co si n tx3 = tx3 o rot3_cs co (~si) n
rotated3 :: !a !.(Vec3 a) v:(!(Vec3 a)-> .b)-> u:(!.(Vec3 a)-> .b) | *,+,-,cos,sin,~ a,[u<=v]
rotated3 ang n tx3 = tx3 o rot3_cs co si n
where
  co=cos ang
  si=(~(sin ang))
chboard3 :: !(Vec3 !.Real) -> Real
chboard3 (V3 x y z)
 # k=entier(2.0*z)
 # n=entier(2.0*y)
 # m=entier(2.0*x)
 = toReal ((n pmod 2 + m pmod 2 + k pmod 2) mod 2)







