implementation module CTColour

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Basic colour routines, etc.

import CTGeneral,StdEnv
from StdPictureDef import Colour,RGBColour,RGB


// Real colours (or, more than reals. I keep it generic in order
//   to compute gradients one day. 

:: RealCol a = {tr::a,tg::a,tb::a}
:: RealColour :== RealCol !Real
:: RColour :== RealCol !Real

instance zero (RealCol a) | zero a
where
 zero = {tr=zero,tg=zero,tb=zero}
instance one (RealCol a) | one a
where
 one = {tr=one,tg=one,tb=one}
 
instance fmap RealCol where
 fmap f {tr=r,tg=g,tb=b} = {tr=f r,tg=f g,tb=f b}
instance *> RealCol where
 (*>) x s = fmap ((*) x) s

instance >/ RealCol where
 (>/) {tr=r,tg=g,tb=b} x = {tr=r/x,tg=g/x,tb=b/x}
 
instance + (RealCol a) | + a
where
 (+) {tr=r,tg=g,tb=b} {tr=x,tg=y,tb=z} = {tr=r+x,tg=g+y,tb=b+z}
instance - (RealCol a) | - a
where
 (-) {tr=r,tg=g,tb=b} {tr=x,tg=y,tb=z} = {tr=r-x,tg=g-y,tb=b-z}
instance * (RealCol a) | * a
where
 (*) {tr=r,tg=g,tb=b} {tr=x,tg=y,tb=z} = {tr=r*x,tg=g*y,tb=b*z}
instance / (RealCol a) | / a
where
 (/) {tr=r,tg=g,tb=b} {tr=x,tg=y,tb=z} = {tr=r/x,tg=g/y,tb=b/z}

instance == (RealCol a) | == a
where
 (==) {tr=r,tg=g,tb=b} {tr=x,tg=y,tb=z} = r==x && g==y && b==z

instance <+> RealCol where
 (<+>) a b = a+b
instance <-> RealCol where
 (<->) a b = a-b
instance *>> (RealCol Real) where
 (*>>) s {tr=r,tg=g,tb=b} = {tr=s*r,tg=s*g,tb=s*b}
 
// To standard system RGB and vice-versa
convertCol :: !(RealCol !Real) -> !Colour
convertCol {tr=a,tg=b,tb=c}
 = RGB {r=entier (255.999999*a),
        g=entier (255.999999*b),b=entier (255.999999*c)}
iconvCol :: !Colour -> !RColour 
iconvCol (RGB {r,g,b})
 = {tr=toReal r/256.0,tg=toReal g/256.0,tb=toReal b/256.0}
 
blackRGB :: !Colour
blackRGB =: RGB {r=0,g=0,b=0}
whiteRGB :: !Colour
whiteRGB =: RGB {r=255,g=255,b=255}
redRGB :: !Colour
redRGB =: RGB {r=255,g=0,b=0}
greenRGB :: !Colour
greenRGB =: RGB {r=0,g=255,b=0}
blueRGB :: !Colour
blueRGB =: RGB {r=0,g=0,b=255}
yellowRGB :: !Colour
yellowRGB =: RGB {r=255,g=255,b=0}
cyanRGB :: !Colour
cyanRGB =: RGB {r=0,g=255,b=255}
magentaRGB :: !Colour
magentaRGB =: RGB {r=255,g=0,b=255}


RRed :: RealCol !Real
RRed  =: {tr=1.0,tg=0.0,tb=0.0}
RPink :: RealCol !Real
RPink =: RRed+0.8*>RCyan
RDarkRed :: RealCol !Real
RDarkRed =: 0.5*>RRed
RLightRed :: RealCol !Real
RLightRed =: RRed + 0.4*>RCyan
RGreen :: RealCol !Real
RGreen  =: {tr=0.0,tg=1.0,tb=0.0}
RDarkGreen :: RealCol !Real
RDarkGreen =: 0.3*>RGreen
RLightGreen :: RealCol !Real
RLightGreen =: RGreen + 0.5*>RMagenta
RBlue :: RealCol !Real
RBlue  =: {tr=0.0,tg=0.0,tb=1.0}
RBlack :: RealCol !Real
RBlack  =: {tr=0.0,tg=0.0,tb=0.0}
RWhite :: RealCol !Real
RWhite  =: {tr=1.0,tg=1.0,tb=1.0}
RCyan :: RealCol !Real
RCyan =: {tr=0.0,tg=1.0,tb=1.0}
RDarkCyan :: RealCol !Real
RDarkCyan=:0.5*>RCyan
RYellow :: RealCol !Real
RYellow =: {tr=1.0,tg=1.0,tb=0.0}
RLightYellow :: RealCol !Real
RLightYellow =: RYellow + 0.4*>RBlue
RMagenta :: RealCol !Real
RMagenta =: {tr=1.0,tg=0.0,tb=1.0}
RGold :: RealCol !Real
RGold =: {tr=0.8,tg=0.5,tb=0.2}
RLightGold :: RealCol !Real
RLightGold =: {tr=1.0,tg=0.72,tb=0.42}
RDarkBlue :: RealCol !Real
RDarkBlue =: {tr=0.0,tg=0.0,tb=0.4}
RLightBlue :: RealCol !Real
RLightBlue=:RBlue+0.6*>RYellow
RGrey :: RealCol !Real
RGrey =: {tr=0.5,tg=0.5,tb=0.5}
RLightGrey :: RealCol !Real
RLightGrey =: {tr=0.8,tg=0.8,tb=0.8}
RDarkGrey :: RealCol !Real
RDarkGrey =: {tr=0.3,tg=0.3,tb=0.3}
RDarkBrown :: RealCol !Real
RDarkBrown=:{tr=0.43,tg=0.23,tb=0.08}
RBrown :: RealCol !Real
RBrown=:{tr=0.6,tg=0.32,tb=0.1}
RLightBrown :: RealCol !Real
RLightBrown=:{tr=0.78,tg=0.58,tb=0.3}

:: CColour a :== (a -> RealCol !Real)
CWhite :: CColour a
CWhite =: const RWhite
CBlack :: CColour a
CBlack =: const RBlack
CRed :: CColour a
CRed =: const RRed
CDarkRed :: CColour a
CDarkRed =: const RDarkRed
CBlue :: CColour a
CBlue =: const RBlue
CDarkBlue :: CColour a
CDarkBlue =: const RDarkBlue
CGreen :: CColour a
CGreen =: const RGreen
CDarkGreen :: CColour a
CDarkGreen =: const RDarkGreen
CYellow :: CColour a
CYellow =: const RYellow
CCyan :: CColour a
CCyan =: const RCyan
CMagenta :: CColour a
CMagenta =: const RMagenta
CGrey :: CColour a
CGrey =: const RGrey
CLightGrey :: CColour a
CLightGrey = const RLightGrey
CDarkGrey :: CColour a
CDarkGrey = const RDarkGrey
CGold :: CColour a
CGold =: const RGold
CBrown :: CColour a
CBrown =: const RBrown

mkgrey :: !Real -> Colour
mkgrey c = RGB {r=cc,g=cc,b=cc} where cc=entier(255.99999*c)

saturmul :: !(RealCol !a) -> !(RealCol a) | zero,==,/,< a
saturmul (cm=:{tr=r,tg=g,tb=b})
 # mx = max3 r g b
 = if (mx<>zero) {tr=(r/mx),tg=(g/mx),tb=(b/mx)} cm

lighten :: !(RealCol !a) -> !(RealCol a) | one,==,+,-,< a
lighten {tr=r,tg=g,tb=b}
 # mx = max3 r g b
 # cx = one-mx
 = {tr=(r+cx),tg=(g+cx),tb=(b+cx)} 

// Luminosity, somewhat about RC_709...
lumin :: !(RealCol !Real) -> Real
lumin {tr,tg,tb} = 0.222*tr+0.707*tg+0.071*tb
// RGB to HSV. Not very useful
rgbtohsv :: (RealCol !Real) -> (Real,Real,Real)
rgbtohsv c=:{tr,tg,tb}
 # (i,v) = minmax3 tr tg tb
 | v==zero = (zero,zero,zero)
   # s=(v-i)/v
   | s==zero = (zero,s,v)
     # dl=fromReal Pi3/(v-i)
     | tr==v = (normangle (dl*(tg-tb)),s,v)
     | tg==v = (fromReal PiD+dl*(tb-tr),s,v)
     = (two*PiD+dl*(tr-tg),s,v)

// HSV to RGB
hsvtorgb :: !Real !Real !Real -> RealCol Real
hsvtorgb h s v
  # h = normangle h
  | s==0.0 = {tr=v,tg=v,tb=v}
    # h = h/Pi3
    # (i,f) = entfrac h
    // # f = hermite f    // insuff. to equilibr. prim/sec cols...
    # f = 0.5*(1.0-cos(Pi*f))
    # p = v*(1.0-s)
    # q = v*(1.0-s*f)
    # t = v*(1.0-s*(1.0-f))
    | i<0 || i>6 = abort ("hsvtorgb error, i="+++toString i)
    = case i of
       0 -> {tr=v,tg=t,tb=p}
       1 -> {tr=q,tg=v,tb=p}
       2 -> {tr=p,tg=v,tb=t}
       3 -> {tr=p,tg=q,tb=v}
       4 -> {tr=t,tg=p,tb=v}
       5 -> {tr=v,tg=p,tb=q}
       6 -> {tr=v,tg=t,tb=p}


