implementation module CTExamples

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// A collection of examples: just texture functions, without
//   interfacing. Not so general as to include them in the
//   main module set.

import CTexture

// ************* Weaving package ************************
modinterlace :: !(!.Int,!.Int,!.Int) !.Int !.Int -> Bool
modinterlace (rowm,rowth,colm) n m = row (m + n pmod colm)
where row m = m pmod rowm < rowth

simple_intl :: (!.Int -> .(!.Int -> Bool))
simple_intl = modinterlace (2,1,2)           // 1x1
lims :: a -> .(a->a) | *,+,-,<,half,one,zero a
lims w
 #! wh = half*w
 = xpulse (half - wh) (half + wh)

weave :: [!RVec2-> a Real] !.Real .[!RVec2-> a Real] !.Real (!RVec2-> a Real) 
         (!Int -> .(!Int -> .Bool)) !.Real -> (!RVec2-> a Real) | *> a
weave hthr hght vthr wdth bkg loom darkw = wv where
  hor = cyclst hthr
  ver = cyclst vthr
  hmask = lims hght
  vmask = lims wdth
  drmask w
   # wh = half*w
   | darkw==zero = one
   = one - hypulse (half - wh) (half + wh) darkw
  hdark = drmask hght
  vdark = drmask wdth
  svdark n m x
    | x<half
      | loom n (m-1) = one
      = vdark (x+one)
    | loom n (m+1) = one
    = vdark (x-one)
  shdark n m y
    | y<half
      | loom (n-1) m = hdark (y+one)
      = one
    | loom (n+1) m = hdark (y-one)
    = one
  wv p=:(V2 x y)
    # (m,rx)=entfrac x
    # (n,ry)=entfrac y
    | (loom n m)
      | (hmask ry)<>zero = svdark n m rx *> (hor n p)
      | (vmask rx)<>zero = hdark ry *> (ver m p)
      = bkg p
    | (vmask rx)<>zero = shdark n m ry *> (ver m p)
    | (hmask ry)<>zero = vdark rx *> (hor n p)
    = bkg p

// Tartan ****************
tartan_intl :: (!.Int -> .(!.Int -> Bool))
tartan_intl = modinterlace (4,2,4)
mksett :: !.Bool ![(a,.Int)] -> .[.b -> a]   
mksett mirror clist
  #! cl = foldr (\(col,nbr) buf -> repeatn nbr (const col) ++ buf) 
         [] clist 
  | mirror = cl ++ reverse cl
  = cl

// ***********************



// ------------------------------------------------------

// Truchet pattern
quarter :: (Vec2 !Real) -> RealCol Real
quarter p = lint RDarkBlue RGold r
where
  f = xpulse 0.15 0.35 o norm2
  r = f p + f (one-p)

truchet :: !(Vec2 !.Real) -> RealCol Real
truchet (p=:(V2 x y))
 # (ix,rx) = entfrac x
 # (iy,ry) = entfrac y
 | ergxy ix iy >= half = quarter (V2 rx ry)
 = quarter (V2 (one-rx) ry)

// ------------------------------------------------------

// Dithering utilities

// "White noise" modulated by a density. This is random dithering texture.
dither :: .Real !.((Vec2 !Real) -> .Real) (Vec2 .Real) -> RealCol Real
dither sf tx p
 # r = grnoise2 (sf*>p) + 0.5
 | tx p > r = RWhite
 = RBlack
makegreybmap :: !(.Int,.Int,{#.Char}) -> .(.(Vec2 !Real) -> Real)
makegreybmap (w,h,bmap) = lumin o imfun
where
  port = bmapPort w h (-1.0) 1.0
  imfun = bmapToTx port Linear (False,False) bmap RGrey

// ------------------------------------------------------

// Bombing
(|+|) infixl 6  :: !a !a -> a | *,+,- a
(|+|) c1 c2 = c1+c2-c1*c2

bombing :: !.Real (!(.Vec2 !Real) -> a) -> .(!(Vec2 !.Real) -> a) | *,+,- a
bombing dst txobj  = f
where
 f p=:(V2 x y)
   # (ix,rx) = entfrac x
   # (iy,ry) = entfrac y
   = ad3 (ad3 rnp ix rx) iy ry
 rnp ix rx iy ry
  # (jx,jy) = ergxy2_s ix iy
  # offx = rx + jx * dst
  # offy = ry + jy * dst
  = txobj (V2 offx offy)
 ad3 f i r = f (i-one) (r+one) |+| f i r |+| f (i+one) (r-one)

// ------------------------------------------------------

// Warping

warp :: !.Real !.(Vec2 !Real) -> .(!(Vec2 !.Real) -> Vec2 Real)
warp k v=:(V2 vx vy) = solve where
  k2 = 2.0*k
  solve p=:(V2 px py) = l!!3 where  // In typical cases it is enough
    l = iterate solver (0.5*>(p-v))
    solver z=:(V2 x y)  // Newton
      # ee = exp(~k*norm2 z)
      # ek = k2*ee
      # ex = ee*vx
      # ey = ee*vy
      # kx = ek*vx
      # ky = ek*vy
      # aj = (V2 (kx*x-1.0) (kx*y), V2 (ky*x) (ky*y-1.0))
      # del = lineq2 aj (V2 (x-px+ex) (y-py+ey))
      = z+del

warpfold :: (!a->v:(!a->a)) !.[a] u:(!a-> .b)->u:(!a -> .b) | +,-,zero a, [u<=v]
warpfold warper trjlist tx = txdef
where
  l = delta trjlist
  (_,txdef) = foldl op (zero,tx) l
  op (shft,txr) v = (shft+v, transf (trfshift shft (warper v)) txr)


// ------------------------------------------------------
// In 3D

// An *image* projected on a sphere (viewport convention changes signs...)
sphermap :: !.(.RVec2 -> .a) !RVec3 -> .a
sphermap imtx p3
  # (V3 r th phi) = tospher (p3)
  = imtx (V2 (Dpi-phi) (Pi-th))
  
spherlift :: !.(.RVec2 -> .a) !RVec3 .b .c -> .a
// Bump map doesn't need normals nor the vision line
spherlift imtx p3 nrm rpos = sphermap imtx p3

// Read two bitmaps: surface and bumps, construct the bumped texture.
mkbplanet :: !{#.Char} !{#.Char} .RVec3 !.Real *a -> *(w:(!(Vec3 .Real) 
   -> v:(!(Vec3 .Real) -> u:(.b -> RealCol Real))),*a) 
   | FileSystem a, [u<=v,v<=w]
mkbplanet surfname bfname blight bcoeff env
  #! (planet,env) = makebmaptx surfname RWhite (True,True) zero Dpi env
  #! txsurf = spherlift planet
  #! (bumps,env) = makebmaptx bfname RWhite (True,True) zero Dpi env
  #! relief = sphermap bumps
  = (bumped blight (\p -> bcoeff*(one - lumin (relief p))) txsurf, 
     env)

// Escher reptiles
txmotifp3 :: (!RVec2-> RColour) !RColour -> .(!.RVec2-> RColour)
txmotifp3 tx bg = f
where
 redst =: Tricell.red_st
 stvec =: Tricell.st_vec
 p0 =: V2 0.5 (0.5/Sqrt3)
 p1 =: V2 1.0 (1.0/Sqrt3)
 p2 =: V2 1.0 0.0
 co120 =: cos PiD
 si120 =: sin PiD; si240 =: ~si120
 rabout a piv = translated piv (rotated_cs co120 a (translated (~piv) tx))
 obj {tr=r,tg=g,tb=b} = abs(1.0-r)>0.01 || abs(1.0-g)>0.01 || b>0.01 // fuzzy yellow mask
 f p
   # sp = redst p
   # rp = stvec sp
   # c = tx rp
   | obj c = RGreen * c
   # c = rabout si120 p0 rp
   | obj c = RGold * c
   # c = rabout si240 p0 rp
   | obj c = RLightBlue * c
   # c = rabout si120 p1 rp
   | obj c = RGold * c
   # c = rabout si240 p1 rp
   | obj c = RLightBlue * c
   # c = rabout si120 p2 rp
   | obj c = RGold * c
   # c = rabout si240 p2 rp
   | obj c = RLightBlue * c
   # c = tx (stvec (sp + V2 0.0 (-1.0)))
   | obj c = RGreen * c
   # c = tx (stvec (sp + V2 1.0 0.0))
   | obj c = RGreen * c
   = bg

kaleidoscope :: v:(!RVec2 -> .a) -> u:(!.RVec2 -> .a), [u<=v]
kaleidoscope tx = f 
where
  p0 =: V2 0.5 (0.5/Sqrt3)
  p1 =: V2 1.0 (1.0/Sqrt3)
  p2 =: V2 0.5 (0.5*Sqrt3)
  ua =: normalize (p2-p0)
  ub =: normalize (p2-p1)
  uh =: normalize (p1-p0)
  redst =: Tricell.red_st
  stuv =: Tricell.st_uv
  stvec =: Tricell.st_vec
  reflectedIn p0 u p = tx (refl p0 u p)
  f p
    # sp = redst p
    # rp = stvec sp
    # (V2 s t) = sp
    # (V2 u v) = stuv sp
    | s<t
      | u>=0.5 && v<1.0 = tx rp
      | u<0.5
        | v>=0.5 = reflectedIn p0 ua rp
        = rotatedAbout p0 PiD tx rp
      | u<1.0 = reflectedIn p1 ub rp
      = rotatedAbout p1 PiA tx rp
    | u<0.5 = reflectedIn p0 ub rp
    | v<0.5 = rotatedAbout p0 PiA tx rp
    | v>=1.0 = reflectedIn p1 ua rp
    | u>=1.0 = rotatedAbout p1 PiD tx rp
    = reflectedIn p0 uh rp

// Parquet
parquet :: !.Real !.Real v:(!RVec2 -> .a) -> u:(!(Vec2 .Real) -> .a),[u<=v]
parquet a b tx = tx o redcell
where
  a2 =: 2.0*a
  b2 =: 2.0*b
  apb=: a+b
  amb=: a-b
  p0 =: V2 a (0.5*b)
  u  =: (Y2axis-X2axis)>/Sqrt2
  tr =: (b/Sqrt2)*>u
  glidrefl p = refl p0 u p + tr
  redcell (V2 x y)
    # su=(y+x) mod a2
    # di=(y-x) mod b2
    = redmotif (0.5*(su-di)) (0.5*(su+di))
  redmotif x y
    | x<0.0 && y<b = glidrefl (V2 (x+apb) (y+amb))
    | (x>=0.0 && x<a && y<b) = V2 x y
    | x>=a = glidrefl (V2 x y)
    = redmotif (x+b) (y-b)


