definition module CTGeneral

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************

import StdString, StdInt, StdReal, StdClass, StdList, StdEnum, StdArray, StdTuple
import StdMaybe

fR :== fromReal

Pi  :: !Real
Dpi :: !Real
PiH :: !Real
PiT :: !Real
Pi3 :: !Real
Pi4 :: !Real
Pi6 :: !Real
PiD :: !Real
PiA :: !Real
Sqrt2 :: !Real
Sqrt3 :: !Real
Sqrt5 :: !Real
SqrtH :: !Real

degrees :: !Real

class Show a where
 showS :: a String -> String
 show :: a -> String
 
instance Show {#Char}
instance Show Int
instance Show Real
instance Show (a,b) | Show a & Show b
instance Show (a,b,c) | Show a & Show b & Show c
instance Show (a,b,c,d) | Show a & Show b & Show c & Show d

instance Show (Maybe a) | Show a
instance Show Bool

instance fromChar Real

class fmap t :: (a->b) (t a) -> (t b)
class fzip t :: (a a->a) (t a) (t a) -> (t a)

class two a :: !a
instance two Int
instance two Real
class three a :: !a
instance three Int
instance three Real
class four a :: !a
instance four Int
instance four Real
class half a :: !a
instance half Real
class onethird a :: !a
instance onethird Real
class onequart a :: !a
instance onequart Real

class (*>)  infixr 7 t   :: !a (t a) -> (t a) | * a
class (>/)  infixr 7 t   :: (t a) !a -> (t a) | / a
class recip a :: !a -> a 
instance recip Real

sqr :: !a -> a | * a
cube :: !a -> a | * a
signum :: !a -> b | <,==,zero a & one,zero,~ b

iterOp :: (!a !a -> a) !Int !a !a -> a
itOp :: (!a !a -> a) !Int !a !a -> a
dsgn :: !Int->Int
pow :: !a !Int -> a | *,one a
fpow :: !a !Int -> a | *,one a
egcd :: !Int !Int -> (Int,Int,Int)

floor :: !Real -> Real
gfloor :: !a -> b | fromInt b & toReal a
nearest :: .a .a !b -> .a | < , half b
frac :: !Real -> Real
gfrac :: !a -> a | -,fromInt,toReal a

entfrac :: !Real -> (Int,Real)
entfr :: !Real -> !(!Int,!Int,!Real,!Real)
isplit ::!Real -> !(Int,Real,Real)
rsplit :: !.Real !.Real -> !.(!Real,!Real)
(pmod) infix 7  :: !Int !Int -> Int
instance mod Real
imod :: !Int !Real -> !Real
bytes :: !Int -> !(!Int,!Int)
digits :: !Int -> ![!Int]
getint :: !{#Char} !Int -> !Int

cmask :: a .b .b -> .b | == , zero a
normangle :: !Real -> Real
atan2 :: !a !a -> a | +,/,<,==,atan,fromReal,zero a
hermite :: !a -> a | *,-,three, two a
herm :: !a -> !(!a,!a) | *,-,one,three,two a
inpol5 :: a -> a | *,+,-,fromReal a
spline :: !a !a !a !a -> .(!a->a) | *,+,-,/,fromReal,half,one,two a



max3 :: !a !a !a -> a | < a
min3 :: !a !a !a -> a | < a
minmax3 :: !a !a !a -> (a,a) | < a
minmax :: ![!a] -> (a,a) | < a
symdiff :: !a !a -> !a | -,abs a

zipWith :: (.a -> .(.b -> .c)) [.a] [.b] -> [.c]
instance fmap []
instance fzip []
instance *> []
cycle :: .[a] -> [a]
cyclst :: .[a] -> .(!Int -> a)
delta :: [a] -> [a] | - a
instance Show [a] | Show a

instance + [a] | + a
instance - [a] | - a
instance * [a] | * a
instance / [a] | / a
instance ~ [a] | ~ a

map2 :: (.a -> b -> .c) [.a] .[b] -> [[.c]]

class (<.>)  infix 9 t   :: (t a) (t a) -> a | *,+ a
class (<->) infixl 6 t   :: (t a) (t a) -> t a | - a
class (<+>) infixl 6 t   :: (t a) (t a) -> t a | + a
class vzero t   :: t a | zero a
class (*>>) infixr 7 a :: !Real !a -> !a

instance *>> Real


class norm2 t :: (t a) -> a | *,+ a
norm :: (a !b) -> !b | * , + , sqrt b & norm2 a
normalize :: (a !b) -> a b | >/ , norm2 a & * , + , / , sqrt b

axproj :: (a !b) (a !b) -> a b | *,+b & *>,<.> a
perp :: (a !b) (a !b) -> a b | *>,<->,<.> a & *,+,- b
linedist2 :: (a !b) (a !b) -> .((a !b)->.(b,b)) | *,+,-,/ b & *>,<->,<.>, norm2 a

lerp :: (a !b) (a !b) !b -> a b | *,+,-,one b & *>,<+> a
clerp :: (a !b) (a !b) !b -> a b | *,+,-,<,one,zero b & *>,<+> a

:: IMethod = Floor | Nearest | Ceiling | Linear | Hermite | Cubic | Spline
interpol :: (a !b) (a !b) .IMethod !b -> a !b | *>,<+> a & *,+,-,<,half,one,three,two,fromReal b

lint :: a a .Real -> a | *>>,+,- a
cmix :: a a v:(.b -> .Real) -> u:(.b -> a) | *>>,+,- a, [u<=v]

