implementation module CTGeneral

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Constants and other stuff of genral applicability. Simple numerics,
//  some list procedures, etc.

import StdString, StdInt, StdReal, StdClass, StdList, StdEnum, StdArray, StdTuple
import StdMaybe

fR :== fromReal

Pi :: !Real
Pi=: 3.141592653589793238462643383
Dpi :: !Real
Dpi=:2.0*Pi
PiH :: !Real
PiH=:0.5*Pi
PiT :: !Real
PiT=:1.5*Pi
Pi3 :: !Real
Pi3=:Pi/3.0
Pi4 :: !Real
Pi4=:Pi/4.0
Pi6 :: !Real
Pi6=:Pi/6.0
PiD :: !Real
PiD=:Dpi/3.0
PiA :: !Real
PiA=:4.0*Pi/3.0

Sqrt2 :: !Real
Sqrt2=:sqrt 2.0
Sqrt3 :: !Real
Sqrt3=:sqrt 3.0
Sqrt5 :: !Real
Sqrt5=:sqrt 5.0
SqrtH :: !Real
SqrtH=:sqrt 0.5

degrees :: !Real
degrees =: Pi/180.0

// generic "show" function (Conversion to string; alternative to
//   toString
class Show a where
 showS :: a String -> String
 show :: a -> String

instance Show {#Char} where
 show x = x
 showS x s = x+++s
instance Show Int where
 show x = toString x
 showS x s = toString x +++ s
instance Show Real where
 show x = toString x
 showS x s = toString x +++ s

instance Show (a,b) | Show a & Show b
where
 showS (x,y) s = showS "(" (showS x (showS "," (showS y (showS ")" s))))
 show a = showS a ""
instance Show (a,b,c) | Show a & Show b & Show c
where
 showS (x,y,z) s = showS "(" (showS x (showS "," (showS y 
                   (showS "," (showS z (showS ")" s))))))
 show a = showS a ""
instance Show (a,b,c,d) | Show a & Show b & Show c & Show d
where
 showS (x,y,z,w) s = showS "(" (showS x (showS "," (showS y (showS "," (showS z
                   (showS "," (showS w (showS ")" s))))))))
 show a = showS a ""


instance Show (Maybe a) | Show a where
 showS Nothing s = "Nothing" +++ s
 showS (Just x) s = showS "Just " (showS x s)
 show a = showS a ""

instance Show Bool where
 showS False s = "False"+++s
 showS True s = "True"+++s
 show a = showS a ""



// Useful...
instance fromChar Real where
 fromChar c = toReal (toInt c)

// Generic maps
class fmap t :: (a->b) (t a) -> (t b)
class fzip t :: (a a->a) (t a) (t a) -> (t a)

// Symbolic numerical constants
class two a :: !a
instance two Int where
  two=:2
instance two Real where
  two=:2.0
class three a :: !a
instance three Int where
  three=:3
instance three Real where
  three=:3.0
class four a :: !a
instance four Int where
  four=:4
instance four Real where
  four=:4.0  
  
class half a :: !a
instance half Real where
 half =: 0.5
class onethird a :: !a
instance onethird Real where
 onethird =: 0.333333333333333333333
class onequart a :: !a
instance onequart Real where
 onequart =: 0.25

class (*>)  infixr 7 t   :: !a (t a) -> (t a) | * a
class (>/)  infixr 7 t   :: (t a) !a -> (t a) | / a
class recip a :: !a -> a 
instance recip Real where recip x = 1.0/x

sqr :: !a -> a | * a
sqr x = x*x
cube :: !a -> a | * a
cube x = x*x*x

signum :: !a -> b | <,==,zero a & one,zero,~ b
signum x | x>zero = one
         | x==zero = zero
         = ~one

// Linear Iterator
iterOp :: (!a !a -> a) !Int !a !a -> a
iterOp _  0 _ buf = buf
iterOp op n x buf = iterOp op (n-1) x (op x buf)

// Dichotomic iterator for associative operators.   
itOp :: (!a !a -> a) !Int !a !a -> a
itOp _  0 _ buf = buf
itOp op n x buf | isOdd n = itOp op (n-1) x (op x buf)
                = itOp op (n / 2) (op x x) buf

// Parity (spec. signum) of a number
dsgn :: !Int->Int
dsgn x | isEven x = 1
       = (-1)

// (Int) Power by linear iteration
pow :: !a !Int -> a | *,one a
pow x n = iterOp (*) n x one
// (Int) Power by binary iteration
fpow :: !a !Int -> a | *,one a
fpow x n = itOp (*) n x one

// Extended Euclid
egcd :: !Int !Int -> (Int,Int,Int)
egcd u 0 = (1,0,u)
egcd u v =
 let eg u1 u3 v1 0 = (u1,(u*u1-u3) / v,u3)
     eg u1 u3 v1 v3 = let q=u3 / v3
                      in eg v1 v3 (u1-v1*q) (u3-v3*q)
 in eg 1 u 0 v 

floor :: !Real -> Real
floor x = toReal (entier x)
gfloor :: !a -> b | fromInt b & toReal a
gfloor x = fromInt (entier (toReal x))
nearest :: .a .a !b -> .a | < , half b
nearest ia ib t = if (t<=half) ia ib

frac :: !Real -> Real
frac x = x-floor x
gfrac :: !a -> a | -,fromInt,toReal a
gfrac x = x - fromInt (entier (toReal x))

entfrac :: !Real -> (Int,Real)
entfrac x
 # n = entier x
 = (n,x-toReal n)
entfr :: !Real -> !(!Int,!Int,!Real,!Real)
entfr x   
  # n = entier x
  # r = x-toReal n
  = (n,n+1,r,r-1.0)
isplit ::!Real -> !(Int,Real,Real)  // keeps toReal n
isplit x
 # n = entier x
 # xn = toReal n
 = (n,xn,x-xn)

rsplit :: !.Real !.Real -> !.(!Real,!Real)
rsplit x a
 # t = a*floor(x/a)
 = (t,x-t)
 
(pmod) infix 7  :: !Int !Int -> Int 
(pmod) a b
 # r=a mod b
 | r<0 = r+b
 = r
// Modulo in Real domain
instance mod Real where
 (mod) a b = a - b * floor (a/b)
// Another mod, mixed
imod :: !Int !Real -> !Real
imod m r
 # ir = entier r
 # nr = ir mod m
 = r - toReal(ir-nr)

// ==============
bytes :: !Int -> !(!Int,!Int)
bytes n = (n mod 256,n/256)
digits :: !Int -> ![!Int]
digits n
  # (a4,n4)=bytes n
  # (a3,n3)=bytes n4
  # (a2,a1)=bytes n3
  = [a4,a3,a2,a1]
getint :: !{#Char} !Int -> !Int  
getint arrch indx
 = fromChar (arrch.[indx]) + 256*(fromChar (arrch.[indx+1]) + 
     256*(fromChar (arrch.[indx+2]) + 256*fromChar (arrch.[indx+3])))

cmask :: a .b .b -> .b | == , zero a
cmask c u v = if (c<>zero) u v
normangle :: !Real -> Real
normangle a = a mod Dpi

// Atan2. generic. Not sure about the order...
atan2 :: !a !a -> a | +,/,<,==,atan,fromReal,zero a
atan2 y x
 | x==zero = if (y<zero) (fromReal PiT) (fromReal PiH)
 # a=atan(y/x)
 | x<zero = a+(fromReal Pi)
 | y<zero = (fromReal Dpi)+a
 = a



hermite :: !a -> a | *,-,three, two a
hermite x = x*x*(three-two*x)
herm :: !a -> !(!a,!a) | *,-,one,three,two a
herm x = (r,one-r) where
  r=x*x*(three-two*x)
inpol5 :: a -> a | *,+,-,fromReal a 
inpol5 x
 # x2=x*x
 = x2*x*(fromReal 10.0 - fromReal 15.0*x + fromReal 6.0*x2)
spline :: !a !a !a !a -> .(!a->a) | *,+,-,/,fromReal,half,one,two a
spline am a0 a1 a2 = fn where
  c1=a1-a0
  c2=half*(a2-two*c1-a0)
  c3=(a0-c1+two*c2-am)/fromReal 6.0
  fn t = a0 + t*(c1+(t-one)*(c2+c3*(t-two)))



max3 :: !a !a !a -> a | < a
max3 x y z | x>y = max x z
           = max y z
min3 :: !a !a !a -> a | < a
min3 x y z | x<y = min x z
           = min y z

minmax3 :: !a !a !a -> (a,a) | < a
minmax3 x y z | x<y = (min x z,max y z)
              = (min y z,max x z)

minmax :: ![!a] -> (a,a) | < a
minmax [x:q] = nx x x q where
  nx n x [] = (n,x)
  nx n x [y:p] =
   if (y>x) (nx n y p) 
            (if (y<n) (nx y x p) (nx n x p))

symdiff :: !a !a -> !a | -,abs a
symdiff f g = abs (f - g)


// *****************************************************
// List Utilities
// *****************************************************
// Copied from extensions
zipWith :: (.a -> .(.b -> .c)) [.a] [.b] -> [.c]
zipWith op [a:as] [b:bs] = [op a b : zipWith op as bs]
zipWith _ _ _ = []

instance fmap [] where
 fmap f l = map f l
instance fzip []
where
 fzip op v1 v2 = zipWith op v1 v2
instance *> [] where
 (*>) x l = map (\s -> x*s) l

cycle :: .[a] -> [a]
cycle x = l where l=x ++ l
cyclst :: .[a] -> .(!Int -> a)
cyclst l = \k -> l!!(k pmod n)
where n =: length l

delta :: [a] -> [a] | - a
delta l=:[l0:lq] = [l0 : (lq-l)]

instance Show [a] | Show a
 where
  showS [] s = s
  showS [x:l] s = showS "[" (showS x (showseq l s))
   where
    showseq [] s = showS "]" s
    showseq [x:q] s = showS "," (showS x (showseq q s))
  show [] = "[]"
  show x = showS x ""

instance + [a] | + a
where
 (+) l1 l2 = zipWith (+) l1 l2
instance - [a] | - a
where
 (-) l1 l2 = zipWith (-) l1 l2
instance * [a] | * a
where
 (*) l1 l2 = zipWith (*) l1 l2
instance / [a] | / a
where
 (/) l1 l2 = zipWith (/) l1 l2
instance ~ [a] | ~ a
where
 ~ l = map ~ l

// *****************************************************
// Interval, map of a binary function over a grid (list of lists)
interv :: a a .Int -> .[a] | *,+,-,/,fromInt a
interv a b n
  # n1 = n-1
  # fc = (b-a)/fromInt n1
  = map (\k -> fromInt k * fc + a) [0 .. n1]

map2 :: (.a -> b -> .c) [.a] .[b] -> [[.c]]
map2 f ls lt
 = map (\s -> map (f s) lt) ls
// *****************************************************

// Generic scalar product for all kind of vectors (including colours)
class (<.>)  infix 9 t   :: (t a) (t a) -> a | *,+ a
class (<->) infixl 6 t   :: (t a) (t a) -> t a | - a
class (<+>) infixl 6 t   :: (t a) (t a) -> t a | + a
class vzero t   :: t a | zero a
class (*>>) infixr 7 a :: !Real !a -> !a

instance *>> Real where
 (*>>) x y = x*y

// In principle it may import <.>, but there are other possible norms
class norm2 t :: (t a) -> a | *,+ a
norm :: (a !b) -> !b | * , + , sqrt b & norm2 a
norm x = sqrt (norm2 x)
normalize :: (a !b) -> a b | >/ , norm2 a & * , + , / , sqrt b
normalize v = v>/norm v
// Projection on NORMALIZED! axes
axproj :: (a !b) (a !b) -> a b | * , + b & *> , <.> a
axproj n v = (v<.>n) *> n

perp :: (a !b) (a !b) -> a b | *> , <-> , <.> a & * , + , - b
perp n v = v <-> (v<.>n) *> n
linedist2 :: (a !b) (a !b) -> .((a !b)->.(b,b)) | *,+,-,/ b & *>,<->,<.>, norm2 a
linedist2 a b = fn where
 u=b<->a
 d=norm2 u
 fn p
  # q=p<->a
  # t=(q<.>u)/d
  = (t,norm2(q<->t*>u))



// Linear interpolation, simple and constrained (clamped)
lerp :: (a !b) (a !b) !b -> a b | *,+,-,one b & *>,<+> a
lerp v0 v1 t = (one-t)*>v0 <+> t*>v1

clerp :: (a !b) (a !b) !b -> a b | *,+,-,<,one,zero b & *>,<+> a
clerp v0 v1 t | t<zero = v0
              | t>one = v1
              = (one-t)*>v0 <+> t*>v1

// Generic, variant interpolation; 6 different methods. Spline not
//  implemented here, useful for multipoint interpolation

:: IMethod = Floor | Nearest | Ceiling | Linear | Hermite | Cubic | Spline
interpol :: (a !b) (a !b) .IMethod !b -> a !b | *>,<+> a & *,+,-,<,half,one,three,two,fromReal b
interpol ia ib meth t
 = case meth of
  Floor -> ia
  Ceiling -> ib
  Nearest -> nearest ia ib t
  Linear ->  (one-t) *> ia <+>  t *> ib
  Hermite -> let y=t*t*(three-two*t) in  (one-y) *> ia <+>  y *> ib
  Cubic -> let y = half+fromReal 4.0*cube(t-half) in  (one-y) *> ia <+>  y *> ib

lint :: a a .Real -> a | *>>,+,- a
lint v0 v1 t = v0 + t*>>(v1-v0)
cmix :: a a v:(.b -> .Real) -> u:(.b -> a) | *>>,+,- a, [u<=v]
cmix c0 c1 f = \p -> lint c0 c1 (f p)
