implementation module CTMapping

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Basic transformations, deformations, 3D mappings...


import CTBase, CTView
import StdMaybe

// transf, displace... are defined in CTFunalg
 
trfshift :: !a .(!a -> a) !a -> a | + , - a
trfshift v tfun p = v + tfun (p-v)
trftranslate :: !a-> .(x:(!a->a)->u:(w:(!a-> .b)->v:(a-> .b))) | +,- a, [v<=w,v u<=x]
trftranslate p0 = transf o (trfshift p0)

trfturn :: !a v:(!.(Vec2 !a)-> .Vec2 a) -> u:(!.(Vec2 !a) -> .Vec2 a) | *,+,-,cos,sin,~ a, [u<=v]
trfturn ang tfun = rot2cs ca sa o (tfun o rot2cs ca (~sa))
where
  ca=cos ang
  sa=sin ang


whirl :: !a !a !.(Vec2 !a)->.Vec2 a | *,+,/,<,==,atan,cos,exp,fromReal,sin,sqrt,zero,~ a
whirl a f (V2 x y)
  # r2 = x*x+y*y
  # r = sqrt r2
  # nphi = atan2 y x + a*exp(~f*r2)
  = V2 (r*cos nphi) (r*sin nphi)
vortex :: !a !a !a !.(Vec2 !a) -> Vec2 a | *,+,-,/,<,cos,sin,sqrt a
vortex rmin2 dr dph p=:(V2 x y)
 # r2=x*x+y*y
 | r2<rmin2 = p
 # r=sqrt r2
 # cf=x/r
 # sf=y/r
 # df = dph/r
 # cs = cos df
 # ss = sin df
 = sqrt(r2+dr) *> V2 (cf*cs-sf*ss) (sf*cs+cf*ss)
 
lense :: a a a -> .(.(Vec2 a) -> Vec2 a) | *,+,-,/,<,one,sqrt,zero,~ a
lense r n h =
 let r0=r*sqrt(one-h*h)
     f (V2 x y)=let
                 rd=sqrt(x*x+y*y)
                 sal=rd/r
                 cal=sqrt(one-sal*sal)
                 sbe=sal/n
                 cbe=sqrt(one-sbe*sbe)
                 d=(cal-h)*(cbe-cal/n)/(cal*cbe+sal*sbe)
                in if (rd>r0) zero (V2 (~d*x) (~d*y))
 in f

spherepoint :: !.Viewscreen -> .(!.RVec2 -> (.Bool,RVec3,RVec3,RVec3))
spherepoint view  = smap where
  ray = view.ray
  pos = view.cam.pos
  ps2 = norm2 pos
  smap p
    # v = ray p
    # sc = pos<.>v
    # dl = sc*sc+1.0-ps2
    | dl<zero = (False,zero,zero,zero)
    # rpos = (sc + sqrt dl)*>v
    # xx = pos - rpos
    = (True,xx,xx,rpos)

cubepoint :: !.Viewscreen -> .(!.RVec2 -> (Bool,RVec3,RVec3,RVec3))
cubepoint view = cmap where
  dum :: (Bool,RVec3,RVec3,RVec3)
  dum = (False,zero,zero,zero)
  ray = view.ray
  pos = view.cam.pos
  (V3 px py pz) = pos
  (V3 rx ry rz) = fmap tsn pos
  tsn s
    | s>1.0 = 1.0
    | s<(-1.0) = (-1.0)
    = 0.0
  test rs ps pa pb vs va vb
    | rs==0.0 = (False,zero,zero)
    # f=(ps-rs)/vs
    # ca = pa-va*f
    | ca<(-1.0) || ca>1.0 = (False,zero,zero)
    # cb = pb-vb*f
    | cb<(-1.0) || cb>1.0 = (False,zero,zero)
    = (True,ca,cb)
  cmap p
    # (V3 vx vy vz) = fmap (\s -> if (s==0.0) (1.0E-14) s) (ray p)
    # (btx,cy,cz) = test rx px py pz vx vy vz
    | btx 
      # u = V3 rx cy cz
      = (True,u,V3 rx 0.0 0.0,pos-u)
    # (bty,cz,cx) = test ry py pz px vy vz vx
    | bty 
      # u = V3 cx ry cz
      = (True,u,V3 0.0 ry 0.0,pos-u)
    # (btz,cx,cy) = test rz pz px py vz vx vy
    | btz 
      # u = V3 cx cy rz
      = (True,u,V3 0.0 0.0 rz,pos-u)
    = dum

cylinderpoint :: !.Viewscreen -> .(.RVec2 -> (.Bool,RVec3,RVec3,RVec3))
cylinderpoint view = cymap where
  pos = view.cam.pos
  (V3 px py pz) = pos
  ray=view.ray
  rc=px*px+py*py-1.0
  rz=if(pz>1.0) 1.0 (if(pz<(-1.0)) (-1.0) 0.0)
  cymap p
    # v = ray p
    # (V3 vx vy vz)=v
    # b = px*vx+py*vy
    # a = 1.0-vz*vz
    # dl = b*b-a*rc
    | dl<0.0 = (False,zero,zero,zero)
    # ti = (b+sqrt dl)/a
    # rpos = ti*>v
    # xx = pos-rpos
    # (V3 x y z) = xx
    | z<1.0 && z>(-1.0) = (True,xx,normalize(V3 x y 0.0),rpos)
    | rz==0.0 = (False,zero,zero,zero)
    # f = (pz-rz)/vz
    # xc = px-vx*f
    # yc = py-vy*f
    | xc*xc+yc*yc>1.0 = (False,zero,zero,zero)
    # xx = V3 xc yc rz
    = (True,xx,V3 0.0 0.0 rz,pos-xx)




proj_plane :: !.(Vec3 a) !a !.(Vec3 a) v:(!(Vec3 a) -> .b) -> u:(!.(Vec2 a) -> .b) 
            | *,+,-,/,sqrt a,[u<=v]
proj_plane n c up0 tx3 = t2
where
  upar =: (up0<.>n)*>n
  yax =: normalize (up0-upar)
  xax =: yax /\ n
  t2 (V2 x y) = tx3 (x*>xax + y*>yax + c*>n)

bumped :: RVec3 (RVec3-> .Real) .(RVec3 -> .(RVec3 -> .(.a -> b Real))) 
          (Vec3 .Real) (Vec3 .Real) .a -> b Real | *> b
bumped blight bmap tx3 p3 nrp rpos
  # gr = fgrad3 0.002 bmap p3
  # n = normalize (nrp - gr)
  # cf = max zero (blight<.>n)
  = (min 1.0 (cf+0.3)) *> tx3 p3 nrp rpos




