definition module CTMath

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************

import CTBase,StdMisc

:: Dif a = Cst a | Dif a (Dif a)

class dif a :: a -> a
instance dif Real
instance dif Int
dCst :: a->Dif a
dVar :: a->Dif a | one a
instance zero (Dif a) | zero a
instance one (Dif a) | one a
instance fromInt (Dif a) | fromInt a
instance fromReal (Dif a) | fromReal a
instance dif (Dif a) | zero a
instance + (Dif a) | + a
instance ~ (Dif a) | ~ a
instance - (Dif a) | - a
instance * (Dif a) | +,* a

instance *> Dif
instance recip (Dif a) | recip,one,+,*,/,-,~ a
instance / (Dif a) | one,+,*,/,-,~ a
instance exp (Dif a) | exp,*,+ a
instance ln (Dif a) | ln,one,+,*,/,-,~ a
instance sqrt (Dif a) | sqrt,one,+,*,/,-,~,fromReal a

dlift :: [a->a] (Dif a) -> (Dif a) | *,+ a
instance sin (Dif a) | sin a & cos a & ~ a & * a & + a
instance cos (Dif a) | sin a & cos a & ~ a & * a & + a
instance atan (Dif a) | atan,one,+,*,/,-,~ a
instance tan (Dif a) | tan,cos,sin,one,+,*,/,-,~ a

instance == (Dif a) | Eq a
instance < (Dif a) | Ord a

toList :: (Dif a) -> [a]
dshow :: (Dif a) -> [a]

instance toReal (Dif a) | toReal a


// ************************************************

:: Scal a = Dfs a [Scal a]
sCst :: a -> Scal a
sVal :: (Scal a) -> a
sDif :: (Scal a) -> [Scal a]
showVal :: (Scal a) -> (a,[a])
instance fromInt (Scal a) | fromInt a
instance fromReal (Scal a) | fromReal a
instance toReal (Scal a) | toReal a

instance == (Scal a) | Eq a
instance < (Scal a) | Ord a
instance zero (Scal a) | zero a
instance one (Scal a) | one a
instance + (Scal a) | + a
instance - (Scal a) | -,~ a
instance ~ (Scal a) | ~ a
unitV :: Int Int -> [a] | zero,one a
dfVar :: Int Int a -> Scal a | zero a & one a

instance fmap Scal
instance *> Scal

instance * (Scal a) | * a & + a
instance recip (Scal a) | recip,one,+,*,/,-,~ a
instance / (Scal a) | one,+,*,/,-,~,recip a
instance exp (Scal a) | exp,*,+ a
instance ln (Scal a) | ln,one,+,*,/,-,~,recip a
instance sqrt (Scal a) | sqrt,one,+,*,recip,/,-,~,fromReal a
instance sin (Scal a) | sin,cos,~,*,+ a
instance cos (Scal a) | cos,sin,~,*,+ a
instance atan (Scal a) | atan,recip,one,+,*,/,-,~ a


mkdVec2 :: .(Vec2 a) -> .Vec2 (Scal a) | one , zero a
dgrad2 :: !(Scal a) -> .Vec2 (Scal a)
grad2 :: !(Scal a) -> .Vec2 a | zero a


// ************** Specif for Vec2 ******************
:: Dv2 a = Dv2 !a !(Vec2 a)
dvCst :: !a -> .Dv2 a | zero a
dvVal :: !.(Dv2 a) -> a
dvgrad :: !.(Dv2 a) -> Vec2 a
mkdv2 :: !.Int a -> .Dv2 a | one , zero a
mkdvec2 :: !.(Vec2 a) -> Vec2 !.(Dv2 a) | one , zero a

instance fromInt (Dv2 a) | fromInt,zero a
instance fromReal (Dv2 a) | fromReal,zero a
instance toReal (Dv2 a) | toReal a
instance == (Dv2 a) | Eq a
instance < (Dv2 a) | Ord a
instance zero (Dv2 a) | zero a
instance one (Dv2 a) | one,zero a
instance half (Dv2 a) | half,zero a
instance two (Dv2 a) | two,zero a
instance three (Dv2 a) | three,zero a
instance + (Dv2 a) | + a
instance - (Dv2 a) | -,~ a
instance ~ (Dv2 a) | ~ a

instance fmap Dv2
instance *> Dv2
instance *>> (Dv2 Real)

instance * (Dv2 a) | *,+ a
instance recip (Dv2 a) | recip,*,~ a
instance / (Dv2 a) | one,+,*,/,-,~,recip a
instance exp (Dv2 a) | exp,* a
instance ln (Dv2 a) | ln,recip,* a
instance sqrt (Dv2 a) | *,sqrt,half,/ a
instance sin (Dv2 a) | sin,cos,* a
instance cos (Dv2 a) | cos,sin,~,* a
instance atan (Dv2 a) | atan,recip,one,+,*,/,-,~ a
instance abs (Dv2 a) | zero,~,< a

// cplx.
:: Cplex a = (:+) infix 9 !a !a
:: Complex :== Cplex Real
instance == (Cplex a) | == a
instance zero (Cplex a) | zero a
instance one (Cplex a) | zero,one a
instance fromReal (Cplex a) | fromReal,zero a
instance fromInt (Cplex a) | fromInt,zero a
instance toReal (Cplex a) | toReal a

instance ~ (Cplex a) | ~ a
instance + (Cplex a) | + a
instance - (Cplex a) | - a
instance * (Cplex a) | +,-,* a

conjg :: (Cplex a)->Cplex a | ~ a
cabs2 :: (Cplex a)->a | +,* a
instance fmap Cplex
instance *> Cplex
instance >/ Cplex

instance recip (Cplex a) | ~,/,+,* a
instance / (Cplex a) | ~,/,+,*,- a
instance exp (Cplex a) | exp,sin,cos,* a
instance sqrt (Cplex a) | ==,<,zero,sqrt,-,+,*,/,fromInt a


// ***************************************
// Rational numbers.
// ***************************************

:: Ratio a = (:%) infix 9 !a !a
instance zero (Ratio a) | zero,one a
instance one (Ratio a) | one a
reduce :: a a -> Ratio a | Eq,zero,/,gcd a
(:/) infix 7 :: a a -> Ratio a | fromInt,Eq,zero,/,*,gcd,sign,abs a
num :: (Ratio a)->a
den :: (Ratio a)->a
instance fromInt (Ratio a) | fromInt,one a
instance ~ (Ratio a) | ~ a
instance sign (Ratio a) | sign a
instance abs (Ratio a) | abs a
instance + (Ratio a) | ==,one,+,gcd,*,/ a
instance - (Ratio a) | ==,one,-,gcd,*,/ a
instance * (Ratio a) | *,/,gcd a
instance / (Ratio a) | fromInt,*,/,gcd,sign,abs a


contour :: a ((Vec2 .(Dv2 a)) -> .Dv2 a) .(Vec2 a) -> Real | *,+,-,/,one,sqrt,toReal,zero a
bumpshade :: !.(Vec3 a) !.((Vec2 !.(Dv2 b)) -> !.Dv2 a) !.(Vec2 b) -> a 
             | *,+,/,one,sqrt,~ a & one,zero b

lic :: !Int !Real (!RVec2->RVec2) (!RVec2->Real)->(!RVec2->Real)

