implementation module CTMath

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Additional mathematical stuff. Principal contents: automatic
//   differentiation package.

// *****************************************************
// Scalar lazy differentiation package (1-Dim)
// *****************************************************

import CTBase,StdMisc

:: Dif a = Cst a | Dif a (Dif a)

class dif a :: a -> a
instance dif Real
where
 dif _ = zero
instance dif Int
where
 dif _ = zero

dCst :: a->Dif a
dCst x = Cst x
dVar :: a->Dif a | one a
dVar x = Dif x (Cst one)


instance zero (Dif a) | zero a
where
 zero = Cst zero
instance one (Dif a) | one a
where
 one = Cst one
instance fromInt (Dif a) | fromInt a
where
 fromInt x = Cst (fromInt x)
instance fromReal (Dif a) | fromReal a
where
 fromReal x = Cst (fromReal x)

instance dif (Dif a) | zero a
where
 dif (Cst _) = zero
 dif (Dif _ p) = p
 
instance + (Dif a) | + a
where
 (+) (Cst x) (Cst y) = Cst (x+y)
 (+) (Cst x) (Dif y p) = Dif (x+y) p
 (+) (Dif x p) (Cst y) = Dif (x+y) p
 (+) (Dif x p) (Dif y q) = Dif (x+y) (p+q)

instance ~ (Dif a) | ~ a
where
 ~ (Cst x) = Cst (~x)
 ~ (Dif x p) = Dif (~x) (~p)
instance - (Dif a) | - a
where
 (-) (Cst x) (Cst y) = Cst (x-y)
 (-) (Cst x) (Dif y p) = Dif (x-y) p
 (-) (Dif x p) (Cst y) = Dif (x-y) p
 (-) (Dif x p) (Dif y q) = Dif (x-y) (p-q)

instance * (Dif a) | +,* a
where
 (*) (Cst x) (Cst y) = Cst (x*y)
 (*) (Cst x) (Dif y p) = Dif (x*y) (x*>p)
 (*) (Dif y p) (Cst x) = Dif (x*y) (x*>p)
 (*) e=:(Dif x p) f=:(Dif y q) = Dif (x*y) (p*f+q*e)

instance *> Dif
where
 (*>) x (Cst y) = Cst (x*y)
 (*>) x (Dif y p) = Dif (x*y) (x*>p)
 
instance recip (Dif a) | recip,one,+,*,/,-,~ a
where
 recip (Cst x) = Cst (recip x)
 recip (Dif x p) = let ip = Dif (recip x) (~p*ip*ip) in ip
instance / (Dif a) | one,+,*,/,-,~ a
where
 (/) (Cst x) (Cst y) = Cst (x/y)
 (/) (Dif x p) f=:(Cst y) = Dif (x/y) (p/f)
 (/) (Cst x) (Dif y p) = let ip = Dif (one/y) (~p*ip*ip)
                         in x*>ip
 (/) e=:(Dif x p) f=:(Dif y q) = Dif (x/y) (p/f - e*q/(f*f))

instance exp (Dif a) | exp,*,+ a
where
 exp (Cst x) = Cst (exp x)
 exp (Dif x p) = let r = Dif (exp x) (p*r)
                 in r
instance ln (Dif a) | ln,one,+,*,/,-,~ a
where
 ln (Cst x) = Cst (ln x)
 ln e=:(Dif x p) = Dif (ln x) (p/e)

instance sqrt (Dif a) | sqrt,one,+,*,/,-,~,fromReal a
where
 sqrt (Cst x) = Cst (sqrt x)
 sqrt e=:(Dif x p) = let r = Dif (sqrt x) (fromReal 0.5*p/r)
                     in r

dlift :: [a->a] (Dif a) -> (Dif a) | *,+ a
dlift [f:fq] e=:(Dif x p) =
  Dif (f x) (p * dlift fq e)

instance sin (Dif a) | sin a & cos a & ~ a & * a & + a
where
 sin (Cst x) = Cst (sin x)
 sin e = dlift (cycle [sin,cos,(~ o sin),(~ o cos)]) e
instance cos (Dif a) | sin a & cos a & ~ a & * a & + a
where
 cos (Cst x) = Cst (cos x)
 cos e = dlift (cycle [cos,(~ o sin),(~ o cos),sin]) e

instance atan (Dif a) | atan,one,+,*,/,-,~ a
where
 atan (Cst x) = Cst (atan x)
 atan e=:(Dif x p) = Dif (atan x) (p/(one+e*e))
instance tan (Dif a) | tan,cos,sin,one,+,*,/,-,~ a
where
 tan (Cst x) = Cst (tan x)
 tan e=:(Dif x p) = let z = cos e in Dif (tan x) (p/(z*z))
// ===============================

instance == (Dif a) | Eq a
where
 (==) (Cst x) (Cst y) = x==y
 (==) (Dif x p) (Dif y q) = x==y
 (==) alpha beta = False
instance < (Dif a) | Ord a
where
 (<) (Cst x) (Cst y) = x<y
 (<) (Cst x) (Dif y _) = x<y
 (<) (Dif x _) (Cst y) = x<y
 (<) (Dif x _) (Dif y _) = x<y

// =============================
prmax = 10

toList :: (Dif a) -> [a]
toList (Cst x) = [x]
toList (Dif x p) = [x:toList p]
dshow :: (Dif a) -> [a]
dshow dd = take prmax (toList dd)

instance toReal (Dif a) | toReal a
where
 toReal (Cst x) = toReal x
 toReal (Dif x _) = toReal x


// *****************************************************************
// N-Dimensional generalization.
// Scalar Sector Only!
// *****************************************************************

:: Scal a = Dfs a [Scal a]
sCst :: a -> Scal a
sCst x = Dfs x []
sVal :: (Scal a) -> a
sVal (Dfs x _) = x
sDif :: (Scal a) -> [Scal a]
sDif (Dfs _ l) = l

showVal :: (Scal a) -> (a,[a])
showVal (Dfs x l) = (x,map sVal l)

instance fromInt (Scal a) | fromInt a
where
 fromInt x = Dfs (fromInt x) []
instance fromReal (Scal a) | fromReal a
where
 fromReal x = Dfs (fromReal x) []
instance toReal (Scal a) | toReal a
where
 toReal (Dfs x _) = toReal x

instance == (Scal a) | Eq a
where
 (==) (Dfs x []) (Dfs y []) = x==y
 (==) (Dfs x p) (Dfs y q) = x==y && (map sVal p)==(map sVal q)

instance < (Scal a) | Ord a     // SILLY !!!!!
where
 (<) (Dfs x _) (Dfs y _) = x<y
 
instance zero (Scal a) | zero a
where
 zero = Dfs zero []
instance one (Scal a) | one a
where
 one = Dfs one []
instance + (Scal a) | + a
where
 (+) (Dfs x []) (Dfs y q) = Dfs (x+y) q
 (+) (Dfs x p) (Dfs y []) = Dfs (x+y) p
 (+) (Dfs x p) (Dfs y q) = Dfs (x+y) (p+q)
instance - (Scal a) | -,~ a
where
 (-) (Dfs x []) (Dfs y q) = Dfs (x-y) (map ~ q)
 (-) (Dfs x p) (Dfs y []) = Dfs (x-y) p
 (-) (Dfs x p) (Dfs y q) = Dfs (x-y) (p-q)
instance ~ (Scal a) | ~ a
where
 ~ (Dfs x p) = Dfs (~x) (map ~ p)

unitV :: Int Int -> [a] | zero,one a
unitV n k = repeatn (k-1) zero ++ 
             [one : repeatn (n-k) zero]
dfVar :: Int Int a -> Scal a | zero a & one a
dfVar n k x = Dfs x (unitV n k)

instance fmap Scal where
 fmap f (Dfs x l) = Dfs (f x) (map (fmap f) l)
instance *> Scal where
 (*>) x s = fmap ((*) x) s

instance * (Scal a) | * a & + a
where
 (*) (Dfs x []) p = x *> p
 (*) p (Dfs x []) = x *> p
 (*) e=:(Dfs x p) f=:(Dfs y q) =
     Dfs (x*y) (map ((*) e) q + map (\z->z*f) p)
instance recip (Scal a) | recip,one,+,*,/,-,~ a
where
 recip (Dfs x []) = Dfs (recip x) []
 recip (Dfs x p) = let ip = Dfs (recip x) (map (\z->z*n) p)
                       n  = ~(ip*ip)
                   in ip
instance / (Scal a) | one,+,*,/,-,~,recip a
where
  (/) (Dfs x []) (Dfs y []) = Dfs (x/y) []
  (/) (Dfs x [])  p = x*>recip p
  (/) p (Dfs x []) = recip x *> p
  (/) e f = e*recip f
 
instance exp (Scal a) | exp,*,+ a
where
 exp (Dfs x []) = Dfs (exp x) []
 exp (Dfs x p) = let r = Dfs (exp x) (map (\z->z*r) p) in r
instance ln (Scal a) | ln,one,+,*,/,-,~,recip a
where
 ln (Dfs x []) = Dfs (ln x) []
 ln e=:(Dfs x p) = Dfs (ln x) (map (\z->z/e) p)

instance sqrt (Scal a) | sqrt,one,+,*,recip,/,-,~,fromReal a
where
 sqrt (Dfs x []) = Dfs (sqrt x) []
 sqrt (Dfs x p) = 
   let r = Dfs (sqrt x) (fromReal 0.5 *> (map (\z->z/r) p))
   in r

instance sin (Scal a) | sin,cos,~,*,+ a
where
 sin (Dfs x []) = Dfs (sin x) []
 sin e=:(Dfs x p) = let c=cos e in Dfs (sin x) (map ((*) c) p)

instance cos (Scal a) | cos,sin,~,*,+ a
where
 cos (Dfs x []) = Dfs (cos x) []
 cos e=:(Dfs x p) = let c = ~(sin e) in Dfs (cos x) (map ((*) c) p)

instance atan (Scal a) | atan,recip,one,+,*,/,-,~ a
where
 atan (Dfs x []) = Dfs (atan x) []
 atan e=:(Dfs x p) = let r=recip(one+e*e)
                     in Dfs (atan x) (map ((*) r) p)

mkdVec2 :: .(Vec2 a) -> .Vec2 (Scal a) | one , zero a
mkdVec2 (V2 x y) = V2 (Dfs x [one, zero]) (Dfs y [zero, one])
dgrad2 :: !(Scal a) -> .Vec2 (Scal a)
dgrad2 (Dfs _ [x, y]) = V2 x y
grad2 :: !(Scal a) -> .Vec2 a | zero a
grad2 (Dfs _ [(Dfs x _), (Dfs y _)]) = V2 x y
grad2 (Dfs _ []) = V2 zero zero


// *****************************************************************
// 2-Dimensional specific variant for the scalar sector.
// Only 1-st order gradients, no laziness.
// *****************************************************************

:: Dv2 a = Dv2 !a !(Vec2 a)
dvCst :: !a -> .Dv2 a | zero a
dvCst x = Dv2 x zero
dvVal :: !.(Dv2 a) -> a
dvVal (Dv2 x _) = x
dvgrad :: !.(Dv2 a) -> Vec2 a
dvgrad (Dv2 _ v) = v

mkdv2 :: !.Int a -> .Dv2 a | one , zero a
mkdv2 1 x = Dv2 x (V2 one zero)
mkdv2 2 y = Dv2 y (V2 zero one)
mkdvec2 :: !.(Vec2 a) -> Vec2 !.(Dv2 a) | one , zero a
mkdvec2 (V2 x y) = V2 (Dv2 x (V2 one zero)) (Dv2 y (V2 zero one))

instance fromInt (Dv2 a) | fromInt,zero a
where
 fromInt x = Dv2 (fromInt x) zero
instance fromReal (Dv2 a) | fromReal,zero a
where
 fromReal x = Dv2 (fromReal x) zero
instance toReal (Dv2 a) | toReal a
where
 toReal (Dv2 x _) = toReal x

instance == (Dv2 a) | Eq a
where
 (==) (Dv2 x p) (Dv2 y q) = x==y && p==q

instance < (Dv2 a) | Ord a 
where
 (<) (Dv2 x _) (Dv2 y _) = x<y
 
instance zero (Dv2 a) | zero a
where
 zero = Dv2 zero zero
instance one (Dv2 a) | one,zero a
where
 one = Dv2 one zero
instance half (Dv2 a) | half,zero a
where
 half = Dv2 half zero
instance two (Dv2 a) | two,zero a
where
 two = Dv2 two zero
instance three (Dv2 a) | three,zero a
where
 three = Dv2 three zero


instance + (Dv2 a) | + a
where
 (+) (Dv2 x p) (Dv2 y q) = Dv2 (x+y) (p+q)
instance - (Dv2 a) | -,~ a
where
 (-) (Dv2 x p) (Dv2 y q) = Dv2 (x-y) (p-q)
instance ~ (Dv2 a) | ~ a
where
 ~ (Dv2 x p) = Dv2 (~x) (~p)

instance fmap Dv2 where
 fmap f (Dv2 x v) = Dv2 (f x) (fmap f v)
 
instance *> Dv2 where
 (*>) x (Dv2 s v) = Dv2 (x*s) (x*>v)

instance *>> (Dv2 Real) where
 (*>>) x (Dv2 s v) = Dv2 (x*s) (x*>v)

instance * (Dv2 a) | *,+ a
where
 (*) (Dv2 x p) (Dv2 y q) =
     Dv2 (x*y) (x*>q + y*>p)
instance recip (Dv2 a) | recip,*,~ a
where
 recip (Dv2 x p)
  #! ix=recip x
  = Dv2 ix (~(ix*ix)*>p)
 
instance / (Dv2 a) | one,+,*,/,-,~,recip a
where
  (/) e f = e*recip f
 
instance exp (Dv2 a) | exp,* a
where
 exp (Dv2 x p)
  # s=exp x
  = Dv2 s (s*>p)
instance ln (Dv2 a) | ln,recip,* a
where
 ln (Dv2 x p) = Dv2 (ln x) (recip x *> p)
instance sqrt (Dv2 a) | *,sqrt,half,/ a
where
 sqrt (Dv2 x p)
  # s=sqrt x
  = Dv2 s ((half/s)*>p)

instance sin (Dv2 a) | sin,cos,* a
where
 sin (Dv2 x p) = Dv2 (sin x) ((cos x)*>p)
instance cos (Dv2 a) | cos,sin,~,* a
where
 cos (Dv2 x p) = Dv2 (cos x) (~(sin x)*>p)

instance atan (Dv2 a) | atan,recip,one,+,*,/,-,~ a
where
 atan (Dv2 x p)
  # r=recip(one+x*x)
  = Dv2 (atan x) (r*>p)

instance abs (Dv2 a) | zero,~,< a
where 
 abs z=:(Dv2 x _)
  | x<zero = ~z
  = z

// Complex Numbers
// ***************

:: Cplex a = (:+) infix 9 !a !a
:: Complex :== Cplex Real

instance == (Cplex a) | == a
where
 (==) (x:+y) (a:+b) = x==a && y==b
instance zero (Cplex a) | zero a
where
 zero=zero:+zero
instance one (Cplex a) | zero,one a
where
 one=one:+zero
instance fromReal (Cplex a) | fromReal,zero a
where
 fromReal x = (fromReal x):+zero
instance fromInt (Cplex a) | fromInt,zero a
where
 fromInt x = (fromInt x):+zero
instance toReal (Cplex a) | toReal a
where
 toReal (x:+_) = toReal x
instance ~ (Cplex a) | ~ a
where
 ~ (x:+y)=(~x):+(~y)
instance + (Cplex a) | + a
where
 (+) (x:+y) (a:+b) = (x+a):+(y+b)
instance - (Cplex a) | - a
where
 (-) (x:+y) (a:+b) = (x-a):+(y-b)
instance * (Cplex a) | +,-,* a
where
 (*) (x:+y) (a:+b) = (x*a-y*b):+(x*b+y*a)

conjg :: (Cplex a)->Cplex a | ~ a
conjg (x:+y) = x:+(~ y)
cabs2 :: (Cplex a)->a | +,* a
cabs2 (x:+y) = x*x + y*y

instance fmap Cplex where
 fmap f (x :+ y) = f x :+ f y
instance *> Cplex 
where
 (*>) s (x:+y) = (s*x) :+ (s*y) 
instance >/ Cplex where
 (>/) (x:+y) s = (x/s) :+ (y/s) 
 
instance recip (Cplex a) | ~,/,+,* a
where
 recip z = conjg z>/cabs2 z
instance / (Cplex a) | ~,/,+,*,- a
where
 (/) n d = (n*conjg d)>/cabs2 d

instance exp (Cplex a) | exp,sin,cos,* a
where
 exp (x:+y) = let e=exp x
              in (e*cos y) :+ (e*sin y)
instance sqrt (Cplex a) | ==,<,zero,sqrt,-,+,*,/,fromInt a
where
 sqrt (a:+b) 
   | b==zero && a>=zero = sqrt a :+ zero
   # y = sqrt ((sqrt(a*a+b*b)-a)/(fromInt 2))
   = (b/(fromInt 2*y)) :+ y

// *****************************************************
// Rational numbers. Optimizations from Knuth
// *****************************************************

:: Ratio a = (:%) infix 9 !a !a
instance zero (Ratio a) | zero,one a
where zero=zero :% one
instance one (Ratio a) | one a
where one=one :% one

reduce :: a a -> Ratio a | Eq,zero,/,gcd a
reduce x y | y == zero  = abort "illegal zero arg"
           = (x/d) :% (y/d) with d = gcd x y

(:/) infix 7 :: a a -> Ratio a | fromInt,Eq,zero,/,*,gcd,sign,abs a
(:/) n d = reduce (n*fromInt(sign d)) (abs d)

num :: (Ratio a)->a
num (x:%y) = x
den :: (Ratio a)->a
den (x:%y) = y

instance fromInt (Ratio a) | fromInt,one a
where
 fromInt n = fromInt n :% one

instance ~ (Ratio a) | ~ a
where
 ~(n:%d)=(~n):%d
instance sign (Ratio a) | sign a
where
 sign (n:%_) = sign n
 
instance abs (Ratio a) | abs a
where
 abs(n:%d) = (abs n):%d
instance + (Ratio a) | ==,one,+,gcd,*,/ a
where
 (+) (un:%ud) (vn:%vd) = 
     let d1=gcd ud vd
     in if (d1==one) ((un*vd+ud*vn):%(ud*vd))
        (let t=un*(vd/d1)+vn*(ud/d1)
             d2=gcd t d1
         in (t/d2):%((ud/d1)*(vd/d2)))
instance - (Ratio a) | ==,one,-,gcd,*,/ a
where
 (-) (un:%ud) (vn:%vd) = 
     let d1=gcd ud vd
     in if (d1==one) ((un*vd-ud*vn):%(ud*vd))
        (let t=un*(vd/d1)-vn*(ud/d1)
             d2=gcd t d1
         in (t/d2):%((ud/d1)*(vd/d2)))

instance * (Ratio a) | *,/,gcd a
where
 (*) (un:%ud) (vn:%vd) = let d1=gcd un vd
                             d2=gcd ud vn
                         in ((un/d1)*(vn/d2)):%((ud/d2)*(vd/d1))
instance / (Ratio a) | fromInt,*,/,gcd,sign,abs a
where
 (/) (un:%ud) (vn:%vd) = 
     let d1=gcd un vn
         d2=gcd ud vd
     in (fromInt(sign vn)*(un/d1)*(vd/d2)):%abs((ud/d2)*(vn/d1))
instance recip (Ratio a) | <,zero,~ a
where
 recip (n:%d) | n>=zero = d:%n
              | otherwise = (~d):%(~n)

// ***************************************************************

contour :: a ((Vec2 .(Dv2 a)) -> .Dv2 a) .(Vec2 a) -> Real | *,+,-,/,one,sqrt,toReal,zero a
contour eps f p
 #! difp = mkdvec2 p
 # r0 = f difp
 # off = normalize (dvgrad r0)
 # p1 = p - eps *> off
 # r = f (mkdvec2 p1)
 = xpulse (toReal r) (toReal r0) zero

bumpshade :: !.(Vec3 a) !.((Vec2 !.(Dv2 b)) -> !.Dv2 a) !.(Vec2 b) -> a 
             | *,+,/,one,sqrt,~ a & one,zero b
bumpshade lpos tx p
  # (V2 dx dy) = dvgrad (tx (mkdvec2 p))
  = lpos <.> normalize (V3 (~dx) (~dy) one)

lic :: !Int !Real (!RVec2->RVec2) (!RVec2->Real)->(!RVec2->Real)
lic nn dt v tx = f
where
  c=1.0/(2.0*fromInt nn + 1.0)
  summ _ 0 tmp _ = tmp
  summ d n tmp p
    #! p1 = p + d*>v p
    = summ d (n-1) (tmp + tx p1) p1
  f p
    #! t1 = summ dt nn (tx p) p
    = c * summ (~dt) nn t1 p


