implementation module CTRandom

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Ergodic primitives, noise, etc.

import CTGeneral, CTVector

// Random numbers etc.
// 649657  599479 25781083          
// 6700417 18837001

// Two alternative generators (aucaso) in [0 - 1]

rand0 :: !Int -> (!Real,!Int)
rand0 seed
 # seed = 649657 + seed*18837001
 # r  = seed bitand 2147483647
 = (toReal r/2147483648.0,seed)
rand1 :: !Int -> (!Real,!Int)
rand1 seed
 # seed = 599479 + seed*25781083
 # r  = seed bitand 2147483647
 = (toReal r/2147483648.0,seed)

// rand0, rand1 symmetric [-0.5 - 0.5]
rand0_h :: !Int -> (!Real,!Int)
rand0_h seed
 # seed = 649657 + seed*18837001
 = (toReal seed/4294967294.0,seed)
rand1_h :: !Int -> (!Real,!Int)
rand1_h seed
 # seed = 599479 + seed*25781083
 = (toReal seed/4294967294.0,seed)

// rand0, rand1 symmetric [-1 - 1]
rand0_s :: !Int -> (!Real,!Int)
rand0_s seed
 # seed = 649657 + seed*18837001
 = (toReal seed/2147483648.0,seed)
rand1_s :: !Int -> (!Real,!Int)
rand1_s seed
 # seed = 599479 + seed*25781083
 = (toReal seed/2147483648.0,seed)
 
// Normal distribution using CLT
randn0 :: !.Int -> (Real,Int)
randn0 seed = ransum 12 0.0 seed where
  ransum :: !Int !Real !Int -> !(!Real,!Int)
  ransum 0 t seed = (t,seed)
  ransum n t seed = 
    let (r,s)=rand0_h seed
    in  ransum (n-1) (t+r) s
randn1 :: !.Int -> (Real,Int)
randn1 seed = ransum 12 0.0 seed where
  ransum :: !Int !Real !Int -> !(!Real,!Int)
  ransum 0 t seed = (t,seed)
  ransum n t seed = 
    let (r,s)=rand1_h seed
    in  ransum (n-1) (t+r) s

// "Ergodic generator": any n -> random
ergodic :: !Int -> !Real
ergodic n
  # n = (n<<13) bitxor n
  = toReal((n*(n*n*15731+789221)+1376312589) bitand 2147483647) 
    / 2147483648.0
ergodic_s :: !Int -> !Real   //  in [-1 : 1]
ergodic_s n
  # n = (n<<13) bitxor n
  = toReal (n*(n*n*599479+649657)+1376312589)/2147483648.0

ergxy :: !Int !Int -> !Real   
ergxy ix iy = ergodic (173 + (13*ix) bitxor (37*iy))
ergxy_s :: !Int !Int -> !Real   
ergxy_s ix iy = ergodic_s (173 + (13*ix) bitxor (37*iy))
ergxy2_s :: !Int !Int -> !(!Real,!Real)
ergxy2_s ix iy
  # n = 497 + (79*ix) bitxor (37*iy)
  # n = (n<<13) bitxor n
  =  (toReal (n*(n*n*599479+649657)+1376312589)/2147483648.0,
      toReal (n*(n*n*15731 +789221)+1376312589)/2147483648.0)

rgxy :: .Int .Int a a -> a | *,+,cos,fromReal,sin a
rgxy ix iy a b
  # n = 497 + (79*ix+12) bitxor (37*iy-31)
  # n = (n<<13) bitxor n
  # r = fromReal (toReal (n*(n*n*157031+789221)+1376312589) / 683565275.57421)
  = a*cos r + b*sin r

ergvecxy :: !Int !Int !Real !Real -> Real   // gen. in [-1 - 1]
ergvecxy  ix iy a b
  # n = 497 + (79*ix+12) bitxor (37*iy-31)
  # n = (n<<13) bitxor n
  # n2=n*n
  =  (a * toReal (n*(n2*599479+649657)+312587137) +
      b * toReal (n*(n2*157031+789221)+1376312589)) / 2147483648.0

/*
grnoise2 :: !(Vec2 !Real) -> Real
grnoise2 (V2 x y)
 # (nx,nx1,rx,mx) = entfr x
 # (ny,ny1,ry,my) = entfr y
 # (nrx,nmx) = herm rx
 # (nry,nmy) = herm ry
 = nmx*(nmy*ergvecxy nx ny rx ry + nry*ergvecxy nx ny1 rx my) + 
   nrx*(nmy*ergvecxy nx1 ny mx ry+ nry*ergvecxy nx1 ny1 mx my)
*/

grnoise2 :: !.(Vec2 a) -> a | *,+,-,cos,fromInt,fromReal,one,sin,toReal a
grnoise2 (V2 x y)
 # (nx,nx1,rx,mx) = entf x
 # (ny,ny1,ry,my) = entf y
 # cry = inpol5 ry // hermite ry
 # aa = rgxy nx ny rx ry
 # a = aa + (rgxy nx ny1 rx my - aa) * cry
 # bb = rgxy nx1 ny mx ry
 # b = bb + (rgxy nx1 ny1 mx my - bb) * cry
 = a + (b-a)*(inpol5 rx) // (hermite rx)
where
 entf x   
  # n = entier (toReal x)
  # r = (x-fromInt n)
  = (n,n+1,r,r-one)    

noise2 :: !.(Vec2 a) -> a | *,+,-,<,cos,fromInt,fromReal,one,sin,toReal,zero a
noise2 p = max zero (min one (fromReal 0.78*(grnoise2 p + fromReal 0.6)))

ergxyz :: !Int !Int !Int -> Real
ergxyz ix iy iz = ergodic (311  + (13*ix) bitxor (37*iy) bitxor (61*iz))

ergnoise3 :: !Int !Int !Int -> !(Vec3 !Real)
ergnoise3 ix iy iz
  # n = 497 + (79*ix) bitxor (37*iy) bitxor (97*iz)
  # n = (n<<13) bitxor n
  # n2=n*n
  =  (V3 (toReal (n*(n2*599479   +649657)+1376312589))
         (toReal (n*(n2*15731    +789221)+1376312589))
         (toReal (n*(n2*6700417+18837001)+1376312589))
          >/ 2147483648.0)
ergvecxyz :: !Int !Int !Int !(Vec3 !Real) -> Real
ergvecxyz  ix iy iz vabc
  =  ergnoise3 ix iy iz <.> vabc

grnoise3 :: !(Vec3 !Real) -> Real
grnoise3 (V3 x y z)
 # (nx,px,rx,mx) = entfr x
 # (ny,py,ry,my) = entfr y
 # (nz,pz,rz,mz) = entfr z
 # (nrx,nmx) = herm rx
 # (nry,nmy) = herm ry
 # (nrz,nmz) = herm rz
 # exyz = ergvecxyz nx ny nz (V3 rx ry rz)
 # enyz = ergvecxyz px ny nz (V3 mx ry rz)
 # exnz = ergvecxyz nx py nz (V3 rx my rz)
 # exyn = ergvecxyz nx ny pz (V3 rx ry mz)
 # ennz = ergvecxyz px py nz (V3 mx my rz)
 # enyn = ergvecxyz px ny pz (V3 mx ry mz)
 # exnn = ergvecxyz nx py pz (V3 rx my mz)
 # ennn = ergvecxyz px py pz (V3 mx my mz)
 = nmz*(nmx*(nmy*exyz+nry*exnz)+nrx*(nmy*enyz+nry*ennz)) +
   nrz*(nmx*(nmy*exyn+nry*exnn)+nrx*(nmy*enyn+nry*ennn))

ergnoise2 :: !.Int !.Int -> Vec2 Real
ergnoise2 ix iy
  # n = 497 + (79*ix) bitxor (113*iy)
  # n = (n<<13) bitxor n
  # n2=n*n
  =  (V2 (toReal (n*(n2*599479   +649657)+1376312589))
         (toReal (n*(n2*6700417+18837001)+1376312589))
          >/ 2147483648.0)

vecnoise2 :: !(Vec2 .Real) -> Vec2 Real          
vecnoise2 (V2 x y)
 # (nx,px,fx) = entfr x
 # (ny,py,fy) = entfr y
 # (rx,mx) = herm fx
 # (ry,my) = herm fy
 = mx*>(my*>ergnoise2 nx ny + ry*>ergnoise2 nx py) + 
   rx*>(my*>ergnoise2 px ny + ry*>ergnoise2 px py)
where
 entfr x   
  # n = entier x
  # r = x-toReal n
  = (n,n+1,r)

fractsum :: ((a !Real) -> !b) (a !Real) !Int !Real !Real -> .((a !Real) -> !b) | *>>,+,zero b & *>,<+>,vzero a
fractsum noise offd noct kappa freq0
  = pn zero noct freq0 1.0 vzero where
    fct=0.5^kappa
    pn t 0 _ _ _ p = t
    pn t n fr fc off p = 
      pn (t+fc*>>noise (fr*>p<+>off)) (n-1) (2.0713*fr)
         (fct*fc) (off <+> offd) p

fractsum2 :: (!Int -> (!Real -> (!Real -> (!RVec2 -> !Real))))
fractsum2 = fractsum grnoise2 (V2 8.37 24.71)
//turbulence2 :: (!Int -> (!Real -> (!Real -> (!RVec2 -> !Real))))
//turbulence2 =: fractsum (abs grnoise2) (V2 3.37 14.71)
fractsum3 :: (!Int -> (!Real -> (!Real -> (!RVec3 -> !Real))))
fractsum3 = fractsum grnoise3 (V3 8.2 24.73 (-3.39))
turbulence3 :: (!Int -> (!Real -> (!Real -> (!RVec3 -> !Real))))
turbulence3 = fractsum (abs grnoise3) (V3 8.2 24.73 (-3.39))

vecnoise3 :: !RVec3 -> !RVec3
vecnoise3 (V3 x y z)
 # (nx,px,fx) = entfr x
 # (ny,py,fy) = entfr y
 # (nz,pz,fz) = entfr z
 # (rx,mx) = herm fx
 # (ry,my) = herm fy
 # (rz,mz) = herm fz
 # exyz = ergnoise3 nx ny nz
 # enyz = ergnoise3 px ny nz
 # exnz = ergnoise3 nx py nz
 # exyn = ergnoise3 nx ny pz
 # ennz = ergnoise3 px py nz
 # enyn = ergnoise3 px ny pz
 # exnn = ergnoise3 nx py pz
 # ennn = ergnoise3 px py pz
 = mz*>(mx*>(my*>exyz+ry*>exnz)+rx*>(my*>enyz+ry*>ennz)) +
   rz*>(mx*>(my*>exyn+ry*>exnn)+rx*>(my*>enyn+ry*>ennn))
where
 entfr x   
  # n = entier x
  # r = x-toReal n
  = (n,n+1,r)

// Generic noise (not Real; adapted to AD)
gennoise2 :: !.(Vec2 a) -> a | *,+,-,cos,fromInt,fromReal,one,sin,toReal a
gennoise2 (V2 x y)
 # (nx,nx1,rx,mx) = entf x
 # (ny,ny1,ry,my) = entf y
 # cry = inpol5 ry // hermite ry
 # aa = rgxy nx ny rx ry
 # a = aa + (rgxy nx ny1 rx my - aa) * cry
 # bb = rgxy nx1 ny mx ry
 # b = bb + (rgxy nx1 ny1 mx my - bb) * cry
 = a + (b-a)*(inpol5 rx)
where
 entf x   
   # n = entier (toReal x) 
   # r = (x-fromInt n)
   = (n,n+1,r,r-one)  


frsum :: ((a !b) -> c) (a !b) !.Int !.Real !b -> .((a !b) -> c) | *>>,+,zero c & *>,<+>,vzero a & *,+,fromReal,zero b
frsum noise offd noct kappa freq0
  = pn zero noct freq0 one vzero where
    nxt = fromReal 1.973
    fct=fromReal(0.5^kappa)
    pn t 0 _ _ _ p = t
    pn t n fr fc off p = 
      pn (t+fc*>>noise (fr*>p <+> off)) (n-1) (nxt*fr)
         (fct*fc) (off <+> offd) p

fractnoise2 :: (!.Int->v:(!.Real->u:(!a -> .(!.(Vec2 a)->a)))) 
               | *,*>>,+,-,cos,fromInt,fromReal,one,sin,toReal,zero a,[u<=v]
fractnoise2 = frsum gennoise2 (V2 (fromReal 8.23) (fromReal (-16.871)))
turbulence2 :: (!.Int->v:(!.Real->u:(!a -> .(!.(Vec2 a)->a)))) 
               | *,*>>,+,-,abs,cos,fromInt,fromReal,one,sin,toReal,zero a, [u<=v]
turbulence2 = frsum (abs gennoise2) (V2 (fromReal 4.57) (fromReal (11.43)))



