implementation module CTTile

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Non-cartesian lattice reduction, symmetries.

import CTVector,CTMapping

// Some constant axes useful for P3, P6 etc. groups,for regular hexes...
T2axis :: !RVec2
T2axis =: V2 0.5 (0.5*Sqrt3)
U2axis :: !RVec2
U2axis =: V2 (-0.5) (0.5*Sqrt3)


// Unit Cell --dealing with translations only.
:: UnitCell = {axS::!RVec2,axT::!RVec2,st_dot::!Real,st_cross::!Real,
               norms2::!Real,normt2::!Real,
               norms::!Real,normt::!Real,
               vec_st::!RVec2-> !RVec2,st_vec::!RVec2-> !RVec2,
               red_st::!RVec2-> !RVec2,
               st_uv::!RVec2-> !RVec2,uv_st::!RVec2-> !RVec2,
               reduce::!RVec2-> !RVec2}

mkunitcell :: !.RVec2 !.RVec2 -> !.UnitCell
mkunitcell axS axT = {axS=axS,axT=axT,st_dot=co,st_cross=si,
                      norms2=nrms2,normt2=nrmt2,
                      norms=nrms, normt=nrmt,
                      vec_st=vec_st,
                      st_vec=st_vec,red_st=red_st,
                      st_uv=st_uv,uv_st=uv_st,
                      reduce=reduce}
where
 nrms2=norm2 axS     // will need when introduce the metrics, cov. coords, etc.
 nrmt2=norm2 axT
 nrms=sqrt nrms2; nrmt=sqrt nrmt2
 co=axS<.>axT
 si=skew axS axT
 //s2=si*si
 vec_st p = V2 (skew p axT / si) (skew axS p / si)
 st_vec (V2 s t) = s*>axS+t*>axT
 red_st p = fmap frac (vec_st p)
 // Contra <-> covariant coordinate conversion 
 st_uv (V2 s t) = V2 (s+t*co/nrms2) (s*co/nrmt2+t)
 uv_st (V2 u v)
   # den = 1.0 - co*co/(nrms2*nrmt2)
   = V2 ((u - v*co/nrms2)/den)
        ((v - u*co/nrmt2)/den)
 reduce p = st_vec (red_st p)

// A useful constant cell
Tricell :: !UnitCell
Tricell =: mkunitcell X2axis T2axis

// Primitive P1 reduction, doesn't touch Motif Region
tile_p1 :: !.UnitCell !.(RVec2 -> !.a) !.RVec2 -> !.a
tile_p1 ucell tx p = tx (ucell.reduce p)

// P1 reduction, permitting to reassemble the Motif Region
tileMR_p1 :: !.UnitCell .(!RVec2 -> !.RVec2) .(!RVec2 -> .a) !.RVec2 -> .a
tileMR_p1 ucell motRed tx p
 # stp = ucell.red_st p
 # np = motRed stp
 = tx (ucell.st_vec np)

hextex :: !.UnitCell !.Real !.RVec2 !.RVec2 w:(RVec2-> .a) v:(RVec2-> .a)-> 
          u:(!.RVec2 -> .a), [u<=v,u<=w]
hextex hcell mortarwidth sp0=:(V2 s0 t0) sp1=:(V2 s1 t1) tx bg = htx
where
  stvec = hcell.st_vec; redst = hcell.red_st
  spa = V2 (s1-1.0) t1; spb = V2 s1 (t1-1.0)
  spc = V2 s0 (t0+1.0); spd = V2 (s0+1.0) t0
  rp0=stvec sp0; rp1=stvec sp1; 
  rpa=stvec spa; rpb=stvec spb; rpc=stvec spc; rpd=stvec spd
  iline a b 
    # u = normalize(b-a)
    = (u,skew a u)
  i0a = iline rp0 rpa; ib0 = iline rpb rp0
  ic1 = iline rpc rp1; i1d = iline rp1 rpd; i01 = iline rp0 rp1  
  testl (dir,c) p
    # side = c - skew p dir
    = (side>=0.0, abs side <= mortarwidth)
  fand (b1,m1) (b2,m2) = (b1&&b2,m1,m2)
  htx p
    # sp = redst p
    # redp = stvec sp
    # (newp,mor) = redmotif redp sp
    | mor = bg redp
    = tx newp
  redmotif p sp=:(V2 s t)
    # (ar1,ma,mb) = fand (testl i0a p) (testl ib0 p)
    | ar1 = (p,ma||mb)
    # (ar2,mc,md) = fand (testl ic1 p) (testl i1d p)
    | ar2 = (stvec(V2 (s-1.0) (t-1.0)),mc||md)
    # (ar3,m0) = testl i01 p
    | ar3 = (stvec(V2 s (t-1.0)),m0||ma||mc)
    = (stvec(V2 (s-1.0) t), m0||mb||md)

regular_hex :: v:(y:(!RVec2-> .a)-> u:(x:(!RVec2-> .a)-> w:(!.RVec2-> .a))), [w<=x,u w<=y,u<=v]
regular_hex = htx where
  p0 = Tricell.uv_st (V2 0.5 0.5)
  p1 = Tricell.uv_st (V2 1.0 1.0)
  htx = hextex Tricell (-0.001) p0 p1

brickwall :: !.Real !.Real !.Real-> v:(y:(!RVec2-> .a)-> u:(x:(RVec2-> .a)-> 
             w:(!.RVec2-> .a))), [w<=x,u w<=y,u<=v]
brickwall a b mwidth
  # Sax = a*>X2axis
  # Tax = (0.5*a)*>X2axis + b*>Y2axis
  # bcell = mkunitcell Sax Tax
  # co = bcell.st_dot
  # nrs = bcell.norms2
  # sp0 = V2 (0.5-0.5*co/nrs) 0.5 
  # sp1 = V2 (1.0-0.5*co/nrs) 0.5
  = hextex bcell mwidth sp0 sp1

p3reduce :: v:(!RVec2-> .a)-> u:(!.RVec2-> .a), [u<=v]
p3reduce tx = ttx
where
 stvec =: Tricell.st_vec
 rp0 =: stvec (one>/3.0)
 rp1 =: stvec (2.0*>one>/3.0)
 ttx p
   # q = Tricell.red_st p
   # (V2 s t)=q
   # rq = stvec q
   # (V2 u v) = Tricell.st_uv q
   | u<=0.5 && s<t = rotatedAbout rp0 PiD tx rq
   | v<=0.5 && s>=t = rotatedAbout rp0 PiA tx rq
   | u>1.0 && s>=t = rotatedAbout rp1 PiD tx rq
   | v>1.0 && s<t = rotatedAbout rp1 PiA tx rq
   = tx rq

