implementation module CTUtils

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Various mixed routines, list sorting, etc. Also: some
//  linear plotting tools to test some functions within the
//  package.

import StdEnv, StdPicture, CTView

// Sorting

:: NTree a = NVoid | NNd a (NTree a) (NTree a)
treesort :: ![!a] -> ![!a] | < a
treesort [] = []
treesort l = flat (build l NVoid) [] where
  flat NVoid t = t
  flat (NNd x lft rgt) t = flat lft [x:flat rgt t]
  build [] t = t
  build [x:q] t = build q (inst x t)
  inst x NVoid = NNd x NVoid NVoid
  inst x (NNd y lft rgt) | x<=y = NNd y (inst x lft) rgt
                         = NNd y lft (inst x rgt)

quicksort :: ![!a] -> ![!a] | < a
quicksort l = qsort l [] where
  qsort :: ![a] !*[a] -> !*[a] | < a
  qsort [] t = t
  qsort [x] t = [x:t]
  qsort [x:l] t = partit x l [] [] t
  partit :: !a ![a] ![a] ![a] !*[a] -> *[a] | < a
  partit x [] sm bg t = qsort sm [x:qsort bg t]
  partit x [y:q] sm bg t| y<=x = partit x q [y:sm] bg t
                        = partit x q sm [y:bg] t

normallist :: a a .[a] -> .[a] | *,+,-,/,< a
normallist fmin fmax lst
  # (mn,mx) = minmax lst
  # rng = (fmax - fmin)/(mx-mn)
  = map (\x->(x-mn)*rng + fmin) lst
ufold :: (!a -> .(*b -> *b)) *b ![!a] -> *b
ufold op ubuf [x:q]
  # ubuf = op x ubuf
  = ufold op ubuf q
ufold _ ubuf [] = ubuf

plotline :: .Viewport [.Vec2 Real] -> .(*Picture -> .Picture)
plotline vport l = 
  plotl (map (vport.vpix) l) 
  where
    plotl [a:q] pic
      = ufold (\x p -> drawLineTo x p) (setPenPos a pic) q
plotfun :: .Viewport .Int .Int (RealCol .Real) (Real -> .Real) *Picture -> .Picture
plotfun vport np wd colr f pic
 # pic = setPenColour (convertCol colr) (setPenSize wd pic)
 # xmin = vport.xmin
 # xmax = vport.xmax
 # dlt = (xmax-xmin)/(fromInt np)
 # xl = [xmin,xmin+dlt .. xmax]
 = plotline vport (map (\x ->(V2 x (f x))) xl) pic

plotfuns :: .Viewport ![(.Int,.Int,RealCol .Real,Real -> .Real)] *Picture -> .Picture
plotfuns vport flist pic
 = ufold (\(np,wd,col,f) p -> plotfun vport np wd col f p)
    pic flist


