implementation module CTVector

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Basic vector routines, etc.

import CTGeneral, CTFunalg

// Vectors. Generic stuff only
:: Vec3 a = V3 !a !a !a
:: RVec3 :== Vec3 Real

instance zero (Vec3 a) | zero a
where zero = V3 zero zero zero
instance one (Vec3 a) | one a
where one = V3 one one one
instance half (Vec3 a) | half a
where half = V3 half half half

instance fromReal (Vec3 a) | fromReal a
where
 fromReal x 
   # s=fromReal x
   = V3 s s s

instance ~ (Vec3 a) | ~ a
where ~(V3 a b c) = V3 (~a) (~b) (~c)
instance + (Vec3 a) | + a
where
 (+) (V3 a b c) (V3 x y z) = V3 (a+x) (b+y) (c+z)
instance - (Vec3 a) | - a
where
 (-) (V3 a b c) (V3 x y z) = V3 (a-x) (b-y) (c-z)
instance * (Vec3 a) | * a
where
 (*) (V3 a b c) (V3 x y z) = V3 (a*x) (b*y) (c*z)
instance / (Vec3 a) | / a
where
 (/) (V3 a b c) (V3 x y z) = V3 (a/x) (b/y) (c/z)
instance == (Vec3 a) | == a
where
 (==) (V3 a b c) (V3 x y z) = a==x && b==y && c==z 
instance Show (Vec3 a) | Show a
where
 showS (V3 x y z) s = showS "V3 " (showS x (showS " " (showS y 
                      (showS " " (showS z s)))))
 show v = showS v ""



instance fmap Vec3 where
 fmap f (V3 x y z) = V3 (f x) (f y) (f z)
instance *> Vec3 where
 (*>) s (V3 a b c) = V3 (s*a) (s*b) (s*c)
instance >/ Vec3 where
 (>/) (V3 a b c) s = V3 (a/s) (b/s) (c/s)
instance <.> Vec3 where
  (<.>) (V3 a b c) (V3 x y z) = a*x+b*y+c*z
instance norm2 Vec3
where
 norm2 (V3 x y z) = x*x+y*y+z*z
instance <+> Vec3 where
 (<+>) a b = (+) a b
instance <-> Vec3 where
 (<->) a b = (-) a b
instance vzero Vec3 where vzero = zero
instance *>> (Vec3 Real) where
 (*>>) s (V3 a b c) = V3 (s*a) (s*b) (s*c)
(/\) infix 9 :: !(Vec3 !a) !(Vec3 !a) -> Vec3 !a | -,* a
(/\) (V3 a b c) (V3 x y z) = V3 (b*z-c*y) (c*x-a*z) (a*y-b*x)

X3axis :: Vec3 !Real
X3axis =: V3 1.0 0.0 0.0
Y3axis :: Vec3 !Real
Y3axis =: V3 0.0 1.0 0.0
Z3axis :: Vec3 !Real
Z3axis =: V3 0.0 0.0 1.0

tospher :: !.(Vec3 !a) -> Vec3 a | *,+,/,<,==,acos,atan,fromReal,sqrt,zero a
tospher p=:(V3 x y z)
 | p==zero = zero
 # r=norm p
 # th = acos (z/r)
 # ph = atan2 y x
 = V3 r th ph
fromspher :: !.(Vec3 !a) -> Vec3 a | *,==,cos,sin,zero a
fromspher (V3 r th phi)
  | r==zero=zero
  # rs=r*sin th
  = V3 (rs*cos phi) (rs*sin phi) (r*cos th)

rot3_cs :: !a !a !.(Vec3 !a) !.(Vec3 !a) -> Vec3 a | *,+,- a
rot3_cs co si n v
  # vpar = (v<.>n) *> n
  # vperp = v-vpar
  # v3 = n /\ vperp
  = vpar + co*>vperp + si*>v3

rot3 :: !a -> .(!.(Vec3 !a) -> .(!.(Vec3 !a) -> Vec3 a)) | *,+,-,cos,sin a
rot3 ang = rot3_cs co si
where
 co=cos ang
 si=sin ang
findrot :: !.(Vec3 a) !.(Vec3 a)-> .(!a,!a,!(Vec3 a)) | *,+,-,/,sqrt a
findrot n v
  # u = normalize v
  # ax = u/\n
  # si = norm ax
  = (u<.>n,si,ax>/si)

// =================================================================

:: Vec2 a = V2 !a !a
:: RVec2 :== Vec2 !Real
instance zero (Vec2 a) | zero a
where zero = V2 zero zero
instance one (Vec2 a) | one a
where one = V2 one one
instance half (Vec2 a) | half a
where half = V2 half half
instance fromReal (Vec2 a) | fromReal a
where
 fromReal x 
   # s=fromReal x
   = V2 s s

instance ~ (Vec2 a) | ~ a
where ~(V2 a b) = V2 (~a) (~b)
instance + (Vec2 a) | + a
where
 (+) (V2 a b) (V2 x y) = V2 (a+x) (b+y)
instance - (Vec2 a) | - a
where
 (-) (V2 a b) (V2 x y) = V2 (a-x) (b-y)
// May seem stupid but useful for scaling
instance * (Vec2 a) | * a
where
 (*) (V2 a b) (V2 x y) = V2 (a*x) (b*y)
instance / (Vec2 a) | / a
where
 (/) (V2 a b) (V2 x y) = V2 (a/x) (b/y)
instance == (Vec2 a) | == a
where
 (==) (V2 a b) (V2 x y) = a==x && b==y

instance Show (Vec2 a) | Show a
where
 showS (V2 x y) s = showS "V2 " (showS x (showS " " (showS y s)))
 show v = showS v ""
 
instance fmap Vec2 where
 fmap f (V2 x y) = V2 (f x) (f y)
instance *> Vec2 where
 (*>) s (V2 a b) = V2 (s*a) (s*b)
instance >/ Vec2 where
 (>/) (V2 a b) s = V2 (a/s) (b/s)
instance <.> Vec2 where
 (<.>) (V2 a b)  (V2 x y) = a*x+b*y
instance norm2 Vec2
where
 norm2 (V2 x y) = x*x+y*y
instance <+> Vec2 where
 (<+>) a b = (+) a b
instance <-> Vec2 where
 (<->) a b = (-) a b
instance vzero Vec2 where vzero = zero

instance *>> (Vec2 Real) where
 (*>>) s (V2 a b) = V2 (s*a) (s*b)

X2axis :: Vec2 !Real
X2axis =: V2 1.0 0.0
Y2axis :: Vec2 !Real
Y2axis =: V2 0.0 1.0

angle :: .(Vec2 a) -> a | +,/,<,==,atan,fromReal,zero a
angle (V2 x y) = atan2 y x

// Skew product and decomposition of p in a non-Cart c.s. (u,v)
skew :: !(Vec2 !a) !(Vec2 !a) -> a | * , - a
skew (V2 x0 y0) (V2 x1 y1) = x0*y1-x1*y0
decomp2 :: !.(Vec2 !a) !.(Vec2 !a) !.(Vec2 !a) -> .(a,a) | *,-,/ a
decomp2 u v p = ((skew p v)/d, (skew u p)/d)
 where d = skew u v

rot2cs :: !a !a .(Vec2 !a) -> .Vec2 a | * , + , - a
rot2cs ca sa (V2 x y) = V2 (x*ca-y*sa) (x*sa+y*ca)
rot2 :: !a !.(Vec2 !a) -> !.(Vec2 !a) | * , + , - , cos , sin a
rot2 a p = rot2cs (cos a) (sin a) p
rot90 :: !.(Vec2 !a) -> !.(Vec2 !a) | ~ a
rot90 (V2 x y) = V2 (~y) x

// reflection in a line
refl :: !(Vec2 a) !(Vec2 a) !(Vec2 a) -> !(Vec2 a) | *,+,-,two a
refl p0 u p
  = two*>(p0 + (u<.>(p-p0))*>u) - p
  
toPolar :: !(Vec2 !a) -> !(a,a) | *,+,/,<,==,atan,fromReal,sqrt,zero a
toPolar (V2 x y)
 = (sqrt (x*x+y*y), atan2 y x)

lineq2 :: (.(Vec2 !a),.(Vec2 !a)) .(Vec2 !a) -> .Vec2 a | *,-,/ a
lineq2 (V2 a00 a01,V2 a10 a11) (V2 p q)
  # det = a00*a11-a10*a01
  = V2 ((p*a11-q*a01)/det) ((a00*q-a10*p)/det)

// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

:: Ray a = Ray !(Vec3 !a) !(Vec3 !a)    // (x0,u) ::: x=x0 + t*>u

sscaled :: a .((b a) -> .c) (b a) -> .c | >/ b & / a
sscaled s tx p = tx (p>/s)

ldiffuse :: (a Real) .Real .((a Real) -> b) (a Real) (a Real) -> b | *>> b & <->,<.> ,>/,norm2 a
ldiffuse lpos amb tx3 p3 nrm
 # cf = normalize(lpos<->p3)<.>nrm + one
 # fc = cf + sqr(one-cf)*amb
 = ramp fc *>> tx3 p3

lstToVec3 :: !u:[.a] -> (Vec3 .a,v:[.a]), [u<=v]
lstToVec3 [a,b,c : l] = (V3 a b c,l)
reducevec :: !(Vec2 .Real) -> (Int,Int,.Vec2 Real)
reducevec (V2 x y)
 # (ix,rx) = entfrac x
 # (iy,ry) = entfrac y
 = (ix,iy,V2 rx ry)

fgrad3 :: !a (v:(Vec3 !a) -> !a) u:(Vec3 !a) -> .Vec3 a | +,-,/ a, [u<=v]
fgrad3 eps f p=:(V3 x y z)
  # f0 = f p
  = V3 ((f (V3 (x+eps) y z) - f0)/eps)
       ((f (V3 x (y+eps) z) - f0)/eps)
       ((f (V3 x y (z+eps)) - f0)/eps)
fgrad2 :: !a (v:(Vec2 !a) -> a) u:(Vec2 !a) -> .Vec2 a | +,-,/ a, [u<=v]
fgrad2 eps f p=:(V2 x y)
  #! f0 = f p
  = V2 ((f (V2 (x+eps) y) - f0)/eps)
       ((f (V2 x (y+eps)) - f0)/eps)

fbumpshade :: !(Vec3 .Real) !.Real !(v:RVec2 -> .Real) !u:(Vec2 .Real) -> Real,[u<=v];
fbumpshade lpos=:(V3 lx ly lz) amb tx p
  # (V2 dx dy) = fgrad2 0.004 tx p
  # nrm = sqrt (dx*dx+dy*dy+1.0)
  # cf0 = (lz-lx*dx-ly*dy)/nrm
  = min 1.0 (max 0.0 cf0 + amb)





