implementation module CTView

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Primitive tools for interfacing. This part EXPLICITLY should never
//  need any Windows specificities, apart the DiB structure, which is
//  independent of actual system libraries. It imports however the
//  StdIOBasic module which declares Point2, etc., and uses files.

import CTVector, StdEnv, StdIOBasic, CTColour

:: Texture a = T2 ((Vec2 !a) -> RealColour) | 
               T3 ((Vec3 !a) (Vec3 !a) (Vec3 !a) -> RealColour)
:: Light = {lpos::!RVec3, lcolour::RColour,spotAt::!RVec3,
            perpAtten::!Real-> !Real, distAtten::!Real-> !Real}
:: Viewport     = {width::!Int,height::!Int,
                   xmin::!Real,xmax::!Real,ymin::!Real,ymax::!Real,
                   xscale::!Real,yscale::!Real, vpix::(!RVec2->Point2),
                   vpix_0::(!RVec2->Point2),pixv::(!Point2-> !RVec2),
                   vxix::!Real-> !Int,vyiy::!Real -> !Int,
		   ixvx::!Int-> !Real,iyvy::!Int-> !Real,
                   vyiy_0::!Real -> !Int, iyvy_0::!Int-> !Real,
                   intfract::(!RVec2 -> !(!Int,!Real,!Int,!Real))
                  }

// Rendering rectangle, pixel size independent of the "camera"
mkvport :: !Int !Int !Real !Real !Real !Real -> !Viewport
mkvport w h xi yi xa ya
  = {width=w,height=h,xmin=xi,xmax=xa,ymin=yi,ymax=ya,xscale=xs,yscale=ys,
     vpix=vpix,vpix_0=vpix_0,pixv=pixv,
     vxix=vxix,vyiy=vyiy,ixvx=ixvx,iyvy=iyvy,
     vyiy_0=vyiy_0,iyvy_0=iyvy_0,
     intfract=intfract}  where
  xr =: xa - xi               // range
  yr =: ya - yi
  xs =: toReal w / xr         // scale
  ys =: toReal h / yr
  vxix x = entier((x-xi)*xs)
  vyiy y = entier((ya-y)*ys)
  ixvx ix = xi + toReal ix/xs
  iyvy iy = ya - toReal iy/ys
  vyiy_0 y = entier((y-yi)*ys) 
  iyvy_0 iy = yi + toReal iy/ys
  vpix =: \(V2 x y) -> {x=entier((x-xi)*xs),y=entier((ya-y)*ys)}
  vpix_0 = \(V2 x y) -> {x=entier((x-xi)*xs),y=entier((y-yi)*ys)}
  pixv =: \{x,y} -> V2 (xi + toReal x/xs) (yi + toReal y/ys)
  intfract (V2 x y)
    # (nx,rx)=entfrac((x-xi)*xs)
    # (ny,ry)=entfrac((y-yi)*ys)
    = (nx,rx,ny,ry)

bmapPort :: !.Int !.Int !.Real !.Real -> Viewport
bmapPort w h xmin xmax
 # htow = fromInt h / fromInt w
 = mkvport w h xmin (htow*xmin) xmax (htow*xmax)  
 
// For the moment just one type of Camera. This will be generalized.
// This one is standard, focal. Will create at least:
//       Orthogonal projection
//       Fish-eye.
:: Camera = {pos::!RVec3, look_at::!RVec3,focal::!Real,rad::!Real,up::!RVec3,
             look_dir::!RVec3,axx::!RVec3,axy::!RVec3,
             orig::!RVec3,nrpos::!Real,dist::!Real}

// Camera def. permits to generate the spatial projections of a point to the
//  virtual screen, but not the final pixel coords.
//  Conventionally the projection screen is at distance=1 from the camera!
//  So, dist: from the coord. orig. look_dir is the v.screen normal.
//  rad is the cone radius THEN. orig is (0,0) on this screen. It is supposed that
//  this corresponds to look_at and is the center of viewport.

mkcamera :: !.(Vec3 !Real) !.(Vec3 !Real) !.Real !.(Vec3 !Real) -> .Camera
mkcamera pos look_at focal up
  # nr = normalize (look_at-pos)   // Normal to the viewing screen
  # axy = normalize (perp nr up)
  # axx = axy /\ nr      // Is sign alright??...
  # orig = nr+pos
  = {pos=pos, look_at=look_at, focal=focal, rad=1.0/focal, up=up, look_dir=nr,
     axx=axx,axy=axy,orig=orig,nrpos=nr<.>pos,dist=nr<.>orig}

readDib :: !*File -> (.Bool,(Int,Int,.{#Char}),.File)
readDib file
  # (_, c1,file) = freadc file
  # (ok,c2,file) = freadc file      // read first two bytes
  | not ok || c1<>'B' || c2<>'M'	  // are they "BM"? 
      = (False,noBitmap,file)
  # (_,  fileSize,file)	= freadi file // read file size
  # (_,  _,       file)	= freadi file // skip bfReserved1 & 2
  # (_,  _,       file)	= freadi file // skip bfOffBits
  # (_,  _,       file)	= freadi file // skip biSize
  # (_,  w,       file)	= freadi file // read width
  # (ok1,h,       file)	= freadi file // read height
  # (ok2,         file)	= fseek  file 0 FSeekSet
  | not ok1 || not ok2 = (False,noBitmap,file)
  # (data,file)			= freads file fileSize
  = (True,(w,h,data),file)
where
  noBitmap = (0,0,{})

readBmp :: !{#.Char} *a -> *((!Int,!Int,.{#Char}),*a) | FileSystem a
readBmp fname world
 # (bool,file,world) = fopen fname FReadData world
 | not bool = abort "cannot open BMP file"
 # (fbool,whbmap,file) = readDib file
 | not fbool = abort "cannot read BMP"
 # (cbool,world) = fclose file world
 = (whbmap,world)

writeBmp :: {#.Char} {#.Char} *a -> *a | FileSystem a
writeBmp dib fname world
 # (bool,file,world) = fopen fname FWriteData world
 # file = fwrites dib file
 # (wbool,world) = fclose file world
 = world
 
setPixel :: !*{#Char} !Int !(RealCol !Real) -> !*{#Char}
setPixel a indx {tr=r,tg=g,tb=b} =
  {a & [indx]=ccnv b,[indx+1]=ccnv g,[indx+2]=ccnv r} where
     ccnv :: !Real -> !Char
     ccnv x = toChar (entier (255.999999999*x))

makeDib :: !.Viewport (!(Vec2 !Real) -> RealCol !.Real) -> (!Int,!Int,!.{#Char})
makeDib vport fun2 = floop a 0 0 54 where
  zer4=:[0,0,0,0]
  h=:vport.height
  w=:vport.width
  w3=:3*w
  pad =: w mod 4
  isize =: (w3+pad)*h
  fsize =: isize+54
  head =: [66,77]++digits fsize++zer4++[54,0,0,0,40,0,0,0]++digits w++digits h++
   [1,0,24,0]++zer4++digits isize ++ [35,11,0,0,35,11,0,0]++zer4++zer4
  a :: *{#Char}
  a = {createArray fsize (toChar 0) & [k]=toChar el \\ el<-head & k<-[0..53]}
  iconv = vport.pixv
  floop :: !*{#Char} !Int !Int !Int -> !(!Int,!Int,!*{#Char})
  floop a iy ix ind
    | iy>=h = (w,h,a)
    | ix>=w = floop a (iy+1) 0 (ind+pad)
    # col = fun2 (iconv {x=ix,y=iy})
    # a = setPixel a ind col 
    = floop a iy (ix+1) (ind+3)

aalias :: !Viewport (!(Vec2 !Real) -> !a) -> (!(Vec2 !Real) -> !a) | *>>,+ a 
aalias vport f = fun5 where
 epsx =: 0.4/vport.xscale
 epsy =: 0.4/vport.yscale
 fun5 p=:(V2 x y) = 
     let xp=x+epsx
         xm=x-epsx
         yp=y+epsy
         ym=y-epsy
     in 0.16666666 *>> (2.0 *>> f p + f (V2 xm ym) + f (V2 xp yp) +
                  f (V2 xm yp) + f (V2 xp ym))

// Making a texture out of a bitmap...

getPixel :: !(.Bool,.Bool) !{#.Char} -> v:(.Int -> u:(.Int -> (.Bool,RColour))), [u<=v]
getPixel (cv,ch) bm = gp where
 w=:getint bm 18
 h=:getint bm 22
 w3=:3*w
 pad =: w mod 4
 wid =: w3+pad
 cr c = fromChar c / 256.0
 gp i j
   # i = if cv (i pmod h) i
   # j = if ch (j pmod w) j
   | i<0 || i>=h || j<0 || j>=w = (False,zero)
   # indx=i*wid+3*j+54
   = (True, {tr=cr(bm.[indx+2]),tg=cr(bm.[indx+1]),tb=cr(bm.[indx])})  

bmapToTx :: !.Viewport !.IMethod !(.Bool,.Bool) !{#.Char} (RealCol .Real)-> .(.RVec2 -> RColour)
bmapToTx vport method cycl bm bg = 
 case method of
   Floor -> pixf
   Linear -> lerp
 where
  gp=getPixel cycl bm
  selgp i j
   # (bool,col) = gp i j
   | bool = col
   = bg
  pixf p
    # ({x=j,y=i})=vport.vpix_0 p
    = selgp i j
  lerp p
    # (j,rx,i,ry) = vport.intfract p
    # i1=i+1
    # j1=j+1
    # my=1.0-ry
    = (1.0-rx)*>(my*>selgp i j  + ry*>selgp i1 j) +
            rx*>(my*>selgp i j1 + ry*>selgp i1 j1)

makebmaptx :: !{#.Char} (RealCol .Real) !(.Bool,.Bool) !.Real !.Real *a -> 
              *(.(.RVec2 -> RealCol Real),*a) | FileSystem a
makebmaptx fname bg cycl xmin xmax env
 # ((w,h,bmap),env) = readBmp fname env
 # port = bmapPort w h xmin xmax
 # imfun = bmapToTx port Linear cycl bmap bg
 = (imfun,env)
 
// This is our projection screen. project v finds the vscreen projection of 
//  a 3D point. 

:: Viewscreen      
   ={cam::!Camera, vport::!Viewport,midx::!Real,midy::!Real,
     project::!RVec3 -> !RVec2, ray::!RVec2 -> !RVec3,
     lights::(!RColour,[Light]),background :: !RVec2 -> !RColour}

mkview :: !.(Vec3 Real) !.(Vec3 Real) !.Real !.(Vec3 Real) !.Int !.Int -> .Viewscreen    
mkview pos look_at focal up width height
  # cam = mkcamera pos look_at focal up
  = mkvscreen cam width height

mkvscreen :: !.Camera !.Int !.Int -> .Viewscreen
mkvscreen cam width height = {cam=cam, vport=vp, midx=w/2.0, midy=h/2.0,
   project=project,ray=ray,lights=(RBlack,[]),
   background=const RGrey} 
where
 w =: toReal width
 h =: toReal height
 asp =: h / w                          // Aspect
 xmax =: cam.rad/sqrt(1.0+asp*asp)     // (Diagonal of the rectangle)
 ymax =: asp*xmax
 vp =: mkvport width height (~xmax) (~ymax) xmax ymax
 nr =: cam.look_dir
 pos =: cam.pos
 ax =: cam.axx
 ay =: cam.axy
 project v
   # u=v-pos
   # t=1.0/(v<.>nr-cam.nrpos)
   # pnt = t*>u - nr
   = V2 (pnt<.>ax) (pnt<.>ay)
 ray (V2 x y) = normalize (x*>ax + y*>ay + nr)

addlights :: !.Viewscreen ((Vec2 !Real) -> .RealCol Real) .(RealCol Real) [.Light] -> .Viewscreen
addlights view=:{lights=(a,l)} bkg amb lghts
 # nlights = (amb,l++lghts)
 = {view & lights=nlights,background=bkg}
mklight :: !.(Vec3 Real) !.(Vec3 Real) .(RealCol Real) (Real -> .Real) (Real -> .Real) -> .Light
mklight pos lookat col spfun attfun
 = {lpos=pos,lcolour=col,spotAt=lookat,
    perpAtten=spfun,distAtten=attfun}

lightview :: ((Vec2 !Real) -> RealCol .Real) !.Viewscreen !.Real (Vec3 !.Real) -> .Viewscreen
lightview bg view ambv lpos
 # lgh = mklight lpos Z3axis RWhite one one
 = addlights view bg (ambv*>RWhite) [lgh]


diffuselight :: !.Light (Vec3 !.Real) (Vec3 !.Real) -> Real
diffuselight light p3 nrm
 # dist = light.lpos <-> p3
 # ndist = norm dist
 # ldir = dist >/ ndist
 # atten = light.distAtten ndist * 
           light.perpAtten (~ldir<.>light.spotAt)
 = atten * (ldir<.>nrm)

solidtx :: !.(Viewscreen -> .(RVec2 -> (.Bool,RVec3,RVec3,RVec3))) 
   !.Viewscreen !.Bool .(RVec3-> .(RVec3 RVec3 -> .RColour)) 
   !.RVec2 -> RColour
solidtx mappoint vpoint lighted tx3 p2
  # (bool,p3,nrm,rpos) = mappoint vpoint p2
  | not bool = vpoint.background p2
  # rawcol = tx3 p3 nrm rpos
  | not lighted = rawcol
  # (amb,lg) = vpoint.lights
  # cd = max zero (diffuselight (hd lg) p3 nrm)
  # cf = cd *> rawcol + sqr(one-cd)*>(amb*rawcol)
  = fmap ramp cf





