implementation module CTWindow

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Windows-specific interfacing issues. Build a rendering window,
//  show the bitmap, etc.

import StdEnv, CTView, StdWindow, StdBitmap, osbitmap, StdPStClass, StdMenuElement
import ostoolbox, StdProcess, pictCCall_12, StdPSt, StdControl, StdId, StdMenu
import Notice
import CTMapping 

// Make Viewport window.
mkportwin :: .{#Char} .Viewport (.Picture -> *Picture) -> Window .NilLS .a *(PSt .b)
mkportwin title vport renderer =
 Window title NilLS
  [ WindowViewSize size, WindowClose (noLS closeProcess)
  , WindowLook True (\_ _->renderer), WindowViewDomain {corner1=org,corner2=maxd}
  ]
where
  size = {w=vport.width,h=vport.height}
  org = {x=0,y=0}
  maxd = {x=org.x+size.w,y=org.y+size.h}

//mainwindow :: (a .Void *(PSt .Void)) *World -> .World | Windows a
mainwindow :: (Window NilLS .Void *(PSt .Void)) *World -> *World
mainwindow window world = startIO SDI Void
    (snd o openWindow Void window) [ProcessClose closeProcess] world

drawDib :: !{#.Char} !.Int !.Int !*Picture -> .Picture
drawDib a w h pic
  # sz = (w,h)
  # (hdc, tb) = WinCreateScreenHDC OSNewToolbox
  # (hbmp,tb) = WinCreateBitmap w a hdc tb
  # tb        = WinDestroyScreenHDC (hdc,tb)
  # bitmap = {originalSize=sz,reSize=sz,bitmapContents=a,bitmapHandle=hbmp}
  = drawAt zero (toBitmap bitmap) pic

txwindow :: {#.Char} !Int !Int !Real !Real !Real !Real (!(Vec2 !Real)->RealCol .Real)->.(*World -> .World)
txwindow name ws hs xi yi xa ya tfun 
  = mainwindow (mkportwin name tecport (drawDib bm w h))
where
  tecport :: !Viewport
  tecport = mkvport ws hs xi yi xa ya
  (w,h,bm) = makeDib tecport tfun

texturwin :: {#.Char} .Viewport ((Vec2 !Real) -> RealCol .Real) -> .(*World -> .World)
texturwin name vport tfun
  = mainwindow (mkportwin name vport (drawDib bm w h))
where
  (w,h,bm) = makeDib vport tfun

// =======================================================================
// The stuff for the second interface
// ***********************************************************************

instance Show Void where
 show Void = "Void"
 showS Void l = "Void"+++l


notixt :: [.{#Char}] .{#Char} !*(.a,*PSt .b) -> *(.a,*PSt .b)
notixt l b (ls,pst)
  # nt = (Notice l (NoticeButton b id) [])
  # ((err,Just a),p) = openModalDialog ls nt pst
  = (a,p)
about :: .(*(.a,*PSt .b) -> *(.a,*PSt .b))
about = notixt ["Welcome to Clastic, texture generator in Clean 1.3",
         "Author: Jerzy Karczmarczuk, University of Caen, France.",
         "( karczma@info.unicaen.fr )",
         "",
         "Version: October 2001"] 
         "Yes"
nope :: .(*(.a,*PSt .b) -> *(.a,*PSt .b))
nope = notixt ["Not implemented yet"] "Sorry"

getMstate :: .Id !*(.a,!*PSt .b) -> *(Bool,*(.a,*PSt .b))
getMstate id (ls,pst=:{io})
  # (Just mst,io) = getParentMenu id io
  # (bol,mrk) = getMenuElementMarkState id mst
  = (marked mrk,(ls,{pst & io=io}))

getSqstate :: .Id !*(.a,*PSt .b) -> *(.Bool,*(.a,*PSt .b))
getSqstate id (loc,pst)
  # (Just wst, pst) = accPIO (getParentWindow id) pst
  # (bool, Just l) = getCheckControlSelection id wst
  = (isMember 1 l,(loc,pst))
togglesq :: !.Id [.Id] !*(.a,!*PSt .b) -> *(.a,*PSt .b)
togglesq myid ables st
  # (bstate,(loc,pst)) = getSqstate myid st
  | bstate = (loc,(appPIO (disableControls ables) pst))
  = (loc,(appPIO (enableControls ables) pst))




:: Txobj = {port::Viewport,bmap::{#Char}}

showBitmap :: !*(.a,!*PSt u:(.Int,w:[y:(.Id,u1:Txobj)])) -> *(.a,*PSt v:(Int,x:[z:(Id,u2:Txobj)])),[u1<=u2,y<=z,w<=x,u<=v]
showBitmap (ls,env)
 # (maybeFile,env)	= selectInputFile env
 | isNothing maybeFile = (ls,env)
 # bitmapfile = fromJust maybeFile
 # (maybeBitmap,env)	= openBitmap bitmapfile env
 | isNothing maybeBitmap = (ls,env)
 # bitmap     = fromJust maybeBitmap
 # bitmapsize = getBitmapSize bitmap
 # {w=ww,h=hh} = bitmapsize
 # (nwid,env) = openId env
 # window     = Window "Bitmap" NilLS
				[WindowViewSize	bitmapsize
				 ,WindowLook True (\_ _->drawAt zero bitmap)
				 ,WindowClose (noLS closeActiveWindow)
				 ,WindowId nwid]
 # (errw,env) = openWindow Void window env
 # bm = (fromBitmap bitmap).bitmapContents
 # port = mkvport ww hh 0.0 0.0 1.0 1.0  // dummy values
 # env = addwin nwid port bm env
 = (ls,env)

savbitmap :: !*(.a,!*PSt (b,[(.Id,.Txobj)])) -> *(.a,*PSt (b,[(Id,Txobj)]))
savbitmap st=:(ls,ps=:{io})
 # (mwid,io) = getActiveWindow io
 # ps = {ps & io=io}
 | mwid==Nothing = (ls,ps)
 # actid = fromJust mwid
 # ((n,lst),ps) = accPLoc (\x->(x,x)) ps
 # bmap = findact actid lst
 # (mname,ps) = selectOutputFile "Save BMP file as" "texture.bmp" ps
 | mname==Nothing = (ls,ps)
 # fname = fromJust mname
 # ps = writeBmp bmap fname ps
 = (ls,ps)
findact :: a [(a,.Txobj)] -> {#Char} | == a
findact id [a:q]
   | fst a == id = (snd a).bmap
   = findact id q 

addwin :: .a .Viewport {#.Char} !*(PSt u:(.Int,w:[y:(.a,u1:Txobj)])) -> *PSt v:(Int,x:[z:(.a,u2:Txobj)]),[u1<=u2,y<=z,w<=x,u<=v]  
addwin wid port bmap ps
  # entr = (wid,{port=port,bmap=bmap})
  # ps = appPLoc (\(n,ls) -> (n+1,[entr:ls])) ps
  = ps
  
close :: !*(PSt (.Int,[(.Id,a)])) -> *PSt (Int,[(Id,a)])
close ps=:{io}
  # (mwid,io) = getActiveWindow io
  # ps = {ps & io=io}
  | mwid==Nothing = ps
  # actid = fromJust mwid
  # ((n,lst),ps) = accPLoc (\x->(x,x)) ps
  # (bool,lst) = removeact actid lst
  | not bool = ps
  # ps = appPLoc (\_ -> (n-1,lst)) ps
  = closeActiveWindow ps
removeact :: a u:[w:(a,.b)] -> (.Bool,v:[x:(a,.b)]) | == a, [w<=x,u<=v]
removeact id [a:q]
 | fst a == id = (True,q)
 = (bres, [a : qq]) with (bres,qq) = removeact id q
removeact id [] = (False,[])

fchoices :: ![(.a,.b)] -> [(.a,.(.c -> .c))]
fchoices fset = fc fset where
 fc [] = []
 fc [(funame,f) : rst] = [(funame, id) : fc rst]

inpline :: .String .Id a .Id b -> LayoutControl (:+: .TextControl (:+: .EditControl .EditControl)) .c .d | toString a & toString b
inpline txt id1 v1 id2 v2 = LayoutControl
  (   TextControl txt [ControlWidth (PixelWidth 110)]
  :+: EditControl (toString v1) (ContentWidth "000.0000") 1
         [ControlId id1,ControlPos (RightToPrev,zero)]
  :+: EditControl (toString v2) (ContentWidth "000.0000") 1
         [ControlId id2,ControlPos (RightToPrev,OffsetVector {vx=3,vy=0})]
  ) [ControlHMargin 0 0,ControlVMargin 0 0,ControlItemSpace 0 0,
     ControlPos (BelowPrev,zero)]
inpline3 :: .String .Id a .Id b .Id c -> LayoutControl (:+: .TextControl (:+: .EditControl (:+: .EditControl .EditControl))) .d .e | toString a & toString b & toString c
inpline3 txt id1 v1 id2 v2 id3 v3 = LayoutControl
  (   TextControl txt [ControlWidth (PixelWidth 110)]
  :+: EditControl (toString v1) (ContentWidth "000.0000") 1
         [ControlId id1,ControlPos (RightToPrev,zero)]
  :+: EditControl (toString v2) (ContentWidth "000.0000") 1
         [ControlId id2,ControlPos (RightToPrev,OffsetVector {vx=3,vy=0})]
  :+: EditControl (toString v3) (ContentWidth "000.0000") 1
         [ControlId id3,ControlPos (RightToPrev,OffsetVector {vx=3,vy=0})]
  ) [ControlHMargin 0 0,ControlVMargin 0 0,ControlItemSpace 0 0,
     ControlPos (BelowPrev,zero)]

quit :== closeProcess
togmark :: !.Bool .Id !*(PSt .a) -> *PSt .a
togmark bool aid pst
   | bool = appPIO(markMenuItems [aid]) pst
   = appPIO(unmarkMenuItems [aid]) pst

stbuts :: .(.a -> .(.b -> .(.c -> *(.d,*PSt .e) -> *(.d,*PSt .e)))) .a .b .c -> LayoutControl (:+: .ButtonControl .ButtonControl) .d *(PSt .e)
stbuts mport fset ctrid ids = LayoutControl
    (ButtonControl "Cancel" [ControlFunction (noLS closeActiveWindow)] :+: 
     ButtonControl "OK" [ControlFunction (mport fset ctrid ids)]) 
    [ControlPos (Center,zero),ControlItemSpace 30 4]

isT2 :: !.(Texture a) -> .Bool
isT2 (T2 f) = True
isT2 _      = False
isT3 :: !.(Texture a) -> .Bool
isT3 (T3 f) = True
isT3 _      = False

createdbwin :: .Viewport .{#Char} ((Vec2 !Real) -> RealCol !.Real) !*(.a,*PSt (.Int,[(.Id,.Txobj)])) -> *(.a,*PSt (Int,[(Id,Txobj)]))
createdbwin port title txfun (ls,ps)
  # (w,h,bm) = makeDib port txfun
  # org = {x=0,y=0}
  # maxd = {x=w,y=h}
  # (nwid,ps) = openId ps
  # win = Window title NilLS
     [WindowViewSize {w=w,h=h}, WindowClose (noLS close)
     ,WindowLook True (\_ _->drawDib bm w h)
     ,WindowViewDomain {corner1=org,corner2=maxd}
     ,WindowId nwid]
  # ps = closeActiveWindow ps  
  # ps = addwin nwid port bm ps  
  # (errw,ps) = openWindow Void win ps 
  = (ls,ps)

makeport :: ![(.{#Char},.Texture Real)] .Id !.[Id] !*(.Id,*PSt (.Int,[(.Id,.Txobj)])) -> *(Id,*PSt (Int,[(Id,Txobj)]))
makeport fset cid idx=:[id1:_] st=:(ls,ps)
  # (Just wst, ps) = accPIO (getParentWindow id1) ps
  # outl = getControlTexts (take 6 idx) wst
  # ((xi,yi,xa,ya,ww,hh),(ls,ps)) = rcr outl (ls,ps)
  # port = mkvport ww hh xi yi xa ya
  # strfun = whichf fset wst (idx!!6)
  # xfun = snd strfun
  | not (isT2 xfun) = (ls,ps)
  # (T2 mfun) = xfun
  # (ismark,(ls,ps)) = getMstate ls (ls,ps)
  # mainfun = if ismark (aalias port mfun) mfun
  #! ps = appPIO (setWindowCursor cid BusyCursor) ps
  # funame = fst strfun
  = createdbwin port funame mainfun (ls,ps)
where
  rcr [(_,Just s1),(_,Just s2),(_,Just s3),
       (_,Just s4),(_,Just s5),(_,Just s6)] st
    # (square,st) = getSqstate (idx!!7) st
    | square = ((toReal s2,toReal s2,toReal s4,toReal s4,toInt s6,toInt s6),st)
    = ((toReal s1,toReal s2,toReal s3,toReal s4,toInt s5,toInt s6),st)

whichf :: ![(.a,.b)] .WState .Id -> (.a,.b)
whichf fset wst cid
  # (bool, Just indx) = getPopUpControlSelection cid wst
  = fset!!(indx-1)

v2dialog :: [(.String,.Texture Real)] .Id (.[a],.[b]) .[Id] *(.Id,*PSt (.Int,[(.Id,.Txobj)])) -> *(Id,*PSt (Int,[(Id,Txobj)])) | toString a & toString b
v2dialog fset cid ([xi,xa,yi,ya],[ww,hh])
         idx=:[id1,id2,id3,id4,id5,id6,id7,id8 : idb] st=:(ls,pst) 
 # stdbuts = stbuts makeport fset cid idx
 # contents = TextControl "Standard Viewport parameters: " [] :+:
     inpline "x_min, y_min: " id1 xi id2 yi :+:
     inpline "x_max, y_max: " id3 xa id4 ya :+:
     inpline "width,height: " id5 ww id6 hh :+:
     PopUpControl
      (fchoices fset)
      1
      [ControlPos (Left,OffsetVector {vx=0,vy=5}),ControlId id7] :+:
     CheckControl 
      [("Square? ",Nothing,NoMark,togglesq id8 [id1,id3,id5])] (Columns 1) 
      [ControlPos (RightToPrev,OffsetVector {vx=90,vy=0}),ControlId id8] :+:
     stdbuts
 # dial = Dialog "New Viewport" contents 
    [WindowClose (noLS closeActiveWindow)]
 # ((err,Just loc), pst) = openModalDialog ls dial pst
 = (ls,pst)

vget :: [(.a,.Maybe b)] (.c,.d) -> (.Viewscreen,(.c,.d)) | toInt , toReal b
vget l st=:(ls,ps)
  # (rl,il)=splitAt 11 l
  # poss = map (\(_,Just s)->toReal s) rl
  # hwd =  map (\(_,Just s)->toInt s) il
  # (pos,rst) = lstToVec3 poss
  # (up ,rst) = lstToVec3 rst
  # (lpos,rst)= lstToVec3 rst
  # focal = hd rst
  # amb = rst!!1
  # (h,w) = (hwd!!0,hwd!!1)
  # vpoint = mkview pos zero focal up h w
  # vpoint = lightview CLightGrey vpoint amb lpos
  = (vpoint,(ls,ps))

txdialog :: .(*a -> *([(.String,.Texture Real)],*World)) *a -> .World | Ids a
txdialog txset world
  # (cid,world) = openId world 
  # (ids,world)	= openIds 20 world   // Controls 2D
  # (jds,world) = openIds 20 world   // Controls 3D
  # (funset,world) = txset world
  = startIO MDI (0,[]) (initinterf funset cid ids jds) //[]: list of windows
       [ProcessClose quit] world

initinterf :: [(.{#Char},.Texture Real)] .Id [.Id] [.Id] !*(PSt (.Int,[(.Id,.Txobj)])) -> *PSt (Int,[(Id,Txobj)])
initinterf funset cid ids jds pst
 # (mid,pst)  = openId pst
 # (err1,pst) = openMenu (ids!!19) menu1 pst
 # (err2,pst) = openMenu 0 (menu2 (ids!!19)) pst    // id20 - Alias state
 # (err3,pst) = openMenu 0 menu3 pst
 # (err4,pst) = openMenu 0 menu4 pst
 = pst
where
 menu1 = Menu "&File" 
    (SubMenu "&New Port"
       (MenuItem "&Standard"  [MenuFunction (v2dialog funset cid Defaultpars ids)] :+:
        MenuItem "&3D"        [MenuFunction (v3dialog funset cid Defaultpars3 jds)] )  [] :+:
     MenuItem "&Close"     [MenuFunction (noLS close)] :+:
     MenuSeparator	       [] :+:
     MenuItem "&Save as"   [MenuFunction savbitmap] :+:
     MenuItem "&Quit"      [MenuShortKey 'q',MenuFunction (noLS quit)]
    ) []
 menu2 aid = Menu "&Control" 
    (MenuItem "&Anti-alias"        [MenuMarkState NoMark,MenuId aid,
                                    MenuFunction (tgalias aid)] :+:
     MenuItem "&Background"        [MenuFunction nope] :+:	
     MenuItem "&Load/show Bitmap"  [MenuFunction showBitmap]
    ) []
 menu3 = Menu "&Params"
    (MenuItem "&Params"  [MenuFunction nope]
    ) []
 menu4 = Menu "&Help" 
    (MenuItem "&Help Window"  [MenuFunction nope] :+:	
     MenuItem "&About"        [MenuFunction about] :+:
     MenuItem "&Show State"   [MenuFunction nope]
    ) []
 tgalias aid st
   # (ismark,(ls,pst)) = getMstate aid st
   # act = if ismark "switched off" "activated"
   # txt = ["Anti-alias option "+++act]
   # pst = togmark (not ismark) aid pst
   = notixt txt "OK" (ls,pst)

v3dialog :: [(.String,.Texture Real)] a (.[b],.[c]) .[Id] *(.Id,*PSt (.Int,[(.Id,.Txobj)])) 
            -> *(Id,*PSt (Int,[(Id,Txobj)])) | toString b & toString c
v3dialog fset cid ([cx,cy,cz,ux,uy,uz,lx,ly,lz,foc,amb],[ww,hh])
         idx=:[id1,id2,id3,id4,id5,id6,id7,id8,id9 : idb]  st=:(ls,pst)
 # stdbuts = stbuts m3port fset cid idx
 # contents = TextControl "3D Viewport parameters: " [] :+:
     inpline3 "Camera pos.  (x,y,z): " id1 (toString cx) id2 (toString cy)
              id3 (toString cz) :+:
     inpline3 "Up direction (x,y,z): " id4 (toString ux) id5 (toString uy)
              id6 (toString uz) :+:
     inpline3 "Light pos.  (x,y,z):  " id7 (toString lx) id8 (toString ly)
              id9 (toString lz) :+:
     inpline "width,height: " (idx!!11) ww (idx!!12) hh :+:
     TextControl "Focal:" [ControlPos (Left,zero)] :+:
     EditControl (toString foc) (ContentWidth "000.0000") 1
         [ControlId (idx!!9),ControlPos (RightToPrev,OffsetVector {vx=72,vy=0})] :+:
     TextControl "Mapping:                 " [ControlPos (Left,zero)] :+:
     RadioControl [("Cube",Nothing,id),("Sphere",Nothing,id),("Cylinder",Nothing,id)] 
                  (Rows 1) 1 [ControlId (idx!!13)] :+:
     PopUpControl
      (fchoices fset) 1
      [ControlPos (Left,OffsetVector {vx=0,vy=5}),ControlId (idx!!14)] :+:
     CheckControl 
      [("Light On? ",Nothing,NoMark,id)] (Columns 1) 
      [ControlPos (RightToPrev,OffsetVector {vx=40,vy=0}),ControlId (idx!!15)] :+:
     TextControl "Ambient: " [ControlPos (RightToPrev,OffsetVector {vx=0,vy=0})] :+:
     EditControl (toString amb) (ContentWidth "000.0000") 1
         [ControlPos (RightToPrev,zero),ControlId (idx!!10)] :+:
     stdbuts
 # dial = Dialog "New 3D Viewport" contents 
    [WindowClose (noLS closeActiveWindow)]
 # ((err,Just ls), pst) = openModalDialog ls dial pst
 = (ls,pst)

m3port :: ![(.{#Char},.Texture Real)] .b !.[Id] !*(.Id,*PSt (.Int,[(.Id,.Txobj)])) 
          -> *(Id,*PSt (Int,[(Id,Txobj)]))
m3port fset cid idx=:[id1:_] st=:(ls,ps)
  # (Just wst, ps) = accPIO (getParentWindow id1) ps
  # (pids,idr) = splitAt 13 idx
  # l = getControlTexts pids wst 
  # (vpoint,(ls,ps)) = vget l (ls,ps)
  # port = vpoint.vport
  # strfun = whichf fset wst (idx!!14)
  # xfun = snd strfun
  | not (isT3 xfun) = (ls,ps)
  # (T3 mfun) = xfun
  # (lighted,(ls,ps)) = getSqstate (idx!!15) (ls,ps)
  # (bool,Just indm) = getRadioControlSelection (idx!!13) wst
  # mapfun = maplist!!(indm-1)
  # tfun = solidtx mapfun vpoint lighted mfun
  # (ismark,(ls,ps)) = getMstate ls (ls,ps)
  # mainfun = if ismark (aalias port tfun) tfun
  # funame = fst strfun
  = createdbwin (vpoint.vport) funame mainfun (ls,ps)
where
  maplist = [cubepoint, spherepoint, cylinderpoint]
