module example3

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// More 2D examples. The same interface as previously. but
//   we import an additional module CTExamples with some
//   specific stuff.

import CTExamples

Start :: *World -> *World
Start world = Render world

// Opens a multi-document interface
Render =: txdialog textureset

// Some auxiliary functions

// Perturbed "wood noise"

tdeform = displace (fR 0.15 * frsum vecnoise2 (V2 3.8 (-20.6)) 5 0.8 2.0)
wnoise p 
  # r = 12.0 * noise2 p
  = interpol RDarkBrown RLightBrown Hermite (gfrac r)

wperturbed = tdeform wnoise

// ... and a perturbed band, different parameters

band0 = (cmix RWhite RDarkBlue o xstripe) (xpulse (-0.3) 0.3)
bdeform = displace (fR 0.6 * frsum vecnoise2 (V2 6.8 10.3) 7 0.8 1.5)
band = bdeform band0

// Perturbed high-frequence "grain"

grain0 = cmix RLightBrown RBlack (scaled (V2 0.02 0.3) noise2)
gdeform = displace (fR 0.2 * frsum vecnoise2 (V2 6.8 10.3) 5 1.1 1.0)
grain = gdeform grain0

// Whirl, Photoshop style
rainbow (V2 x y) = hsvtorgb (8.0*x) one one
eddy = transf (whirl 8.0 6.0) rainbow

// "Physical" vortex: sink
fc=:6.0/Pi
polarchk p
 # (r,phi) = toPolar p
 = lint RYellow RRed (chboard (V2 (3.0*r) (fc*phi)))
sink = transf (vortex 0.004 0.8 1.0) polarchk


// Transforming and combining deformers. 
// Shifted lense

ch0 = (sscaled 0.3 o rotated 0.2) chboard
ch1 = cmix RBlue RLightBrown ch0
deformer = id + lense 0.4 2.5 0.05
defboard = trftranslate (V2 0.35 0.2) deformer ch1

// A "celtic" pattern
undisk = (pulse*id) o norm2
 
eddrot n  = trfturn (fromInt n*0.4*Pi) (trfshift X2axis (whirl 9.0 12.0))
eddy5 = foldl (o) id  (map (transf o eddrot) [0 .. 4])
celtic = sscaled 0.7 (cmix RYellow RDarkBlue (eddy5 undisk))

// ************************************

// Warping.

// A simple one, one step (not too big)
tx0 = sscaled 0.4 (cmix RYellow RBlack chboard)
pattern p
  # np = p+(V2 0.2828427 0.2828427)
  # rd = norm np
  # rng = lint RRed RWhite (trainpulse (12.0*rd))
  = lint (tx0 p) rng (step (0.4-rd))
  
v=:V2 0.35 0.25
k=:5.0
warper =: warp k
   
txfun = transf (warper v) pattern

// Complex warp, following a trajectory (iterate small steps)
traj =: [V2 0.1 (-0.1), V2 0.3 0.0, V2 0.5 0.25,  V2 0.6 0.5, V2 0.7 0.7,
         V2 0.7 0.8, V2 0.6 0.8, V2 0.4 0.7, V2 0.2 0.5]

warpfun = warpfold warper traj pattern


// This generates a set of functions (and descriptive strings) for the interface
textureset world
 # funlist = [("Fr. displacement", T2 wperturbed),
              ("Perturbed band", T2 band),
              ("Perturbed grain",T2 grain),
              ("Whirl",T2 eddy),
              ("Sink",T2 sink),
              ("Shifted lense",T2 defboard),
              ("Celtic", T2 celtic),
              ("Simple Warp",T2 txfun),
              ("Traj. warp",T2 warpfun)
             ]
 = (funlist,world)

           
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!









