module example4

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// More 2D and also 3D examples

import CTExamples

Start :: *World -> *World
Start world = Render world

// Opens a multi-document interface
Render =: txdialog textureset

// Some auxiliary functions
// Cut through 3D checkerboard (cubes)

pattern = sscaled 0.5 chboard3
axn = normalize (V3 1.0 1.0 1.0)           // Diagonal vision line
dst =: 0.125*Sqrt3                           // Middle of cubes
up0 = normalize (Y3axis + 0.25*>X3axis)   // just for fun
cubes0 = proj_plane axn dst up0 pattern
cubtex = cmix RDarkGreen RPink cubes0

// The same 'cuboard' shown on a cube, sphere or cylinder
//  (if you wish)
// Rotate the object to have Z == axn
rotcubes
  # (co,si,ax) = findrot axn Z3axis
  = rotated3_cs co si ax (translated (dst*>one) pattern)

cuboard3 p3 nrm rpos = cmix RRed RBlack rotcubes p3
// Use interface to choose your projection. 
// $$ For a cube change focal to 4!
// $$ For cylinder: to 4.5! Switch light on.
// "Marble" sphere (or other 3D mapping)

tbruit p3 = fractsum3 6 0.7 0.8 p3 + 0.55
marbmap =: [(RPink,0.0),(RPink,0.25),(RDarkBlue,0.4),(RPink,0.50),
            (RPink,0.7),(RDarkBlue,0.85),(RPink,0.95),(RPink,1.0)]
marble p3 nrm rpos = colmap marbmap Hermite (tbruit p3)


// Simple bump mapping
snoise0 = sscaled 0.06 (grnoise3 + fR 0.62)
snoise = fR 0.08 * snoise0

// Light source
blight = normalize (V3 (-2.0) (-1.0) 3.0)
bumptest = bumped blight snoise marble

// Earth etc. showing relief
// Light source for bumping. (We like Europe)
elight = normalize (V3 (-2.0) (1.0) 3.0)


// =====================

// A horrible "fractal planet" which doesn't resemble anything.

harsh p3 = ramp (3.5*(0.4 - turbulence3 6 0.9 1.5 p3))
plcolmap=:[(RGreen,0.0),(RGreen,0.2),(RBrown,0.5),
            (RGrey,0.8),(RWhite,0.95),(RWhite,1.0)]
cplanet p3 n r
  # h = harsh p3
  = cmask h (colmap plcolmap Linear h) RBlue
aplanet = bumped blight (\p -> 0.05*harsh p) cplanet

// =====================

// And, finally, a reconstruction of a 2-dim map from a 3D texture, as above.
//  $$ Make the picture rectangular, width = 2*height; x corresponds to phi:(0,2Pi),
//  $  y is theta: 0 to Pi.
// Actually, there is ALMOST NOTHING TO DO.

makesphmap tx3 (V2 phi theta)
  = tx3 (fromspher (V3 one theta phi))

// Fractal planet transformed into planar maps.  
cplanetmap = makesphmap (\p -> cplanet p p p)
harshmap = makesphmap (cmix RBlack RWhite harsh)

// This generates a set of functions (and descriptive strings) for the interface
//  $$ For Earth/Moon bumps choose camera position: (-4,8,3)
textureset world
 # (tellus,world) = mkbplanet "../Images/EarthMap.bmp" "../Images/EarthBump.bmp"
                   elight 0.15 world
 # (selene,world) = mkbplanet "../Images/MoonMap.bmp" "../Images/MoonBump.bmp"
                   elight 0.08 world
                    
 # funlist = [("Rot. cubes cutthrough", T2 cubtex),
              ("Rotated cubes", T3 cuboard3),
              ("Marble",T3 marble),
              ("Bumps",T3 bumptest),
              ("Earth",T3 tellus),
              ("Moon",T3 selene),
              ("Fractal planet",T3 aplanet),
              ("Planet map", T2 cplanetmap),
              ("Planet bumps",T2 harshmap)
             ]
 = (funlist,world)

           
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// //  # (imfun,world) = makebmaptx "../images/zosia2.bmp" RGrey 0.0 1.0 world



