module example6

// *****************************************************
// Clean Texturing Library
// Copyright 2001 Jerzy Karczmarczuk, University of Caen
// Version 01.10
// *****************************************************
// Some classical textures (SynTex, Corel, RenderMan...)

import CTExamples

Start :: *World -> *World
Start world = Render world

// Opens a multi-document interface
Render =: txdialog textureset

// Some auxiliary functions

nn=:5
pepper p
  # t=grnoise2 (150.0*>p)
  = step t
licnoise dt field p
  # t = lic nn dt field pepper p
  = lint RDarkBlue RYellow t

rnoise = sscaled (fR 0.7) gennoise2
txnoise p = lint RBlack RYellow (0.7*rnoise p + 0.6)

gradnoise p = dvgrad (rnoise (mkdvec2 p))
absgrad p = lint RCyan RGold (0.42*norm (gradnoise p))

dstep=:0.003
curlnoise p
  # (V2 a b) = gradnoise p
  = V2 (~b) a
gradlic = licnoise dstep gradnoise
curllic = licnoise dstep curlnoise

// This generates a set of functions (and descriptive strings) for the interface
textureset world
                    
 # funlist = [("Pepper",T2 (cmix RBlack RWhite pepper)),
              ("Scalar Noise",T2 txnoise),
              ("Abs. grad",T2 absgrad),
              ("Gradient: LIC",T2 gradlic),
              ("Curl: LIC",T2 curllic),
              ("Spot",T2 (cmix RDarkRed RWhite spot)),
              ("Spotgrad", T2 spotgrad),
              ("Spotcurl", T2 spotcurl),
              ("Lic Radial",T2 (licnoise dstep radial)),
              ("Lic Azimuthal",T2 (licnoise dstep rotor))
             ]
 = (funlist,world)
        
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// Auxiliaries

spot p
  # t=grnoise2 (2.0*>p)
  = step (t+0.2)
//  = 0.78*(t+0.66)

cf=:14.0
dfrspot field p
  #! v=field p
  # (V2 a b) = v
  #! d = norm v
  #! cfl = 1.0+cf*d
  #! cft = 1.0/cfl
  #! co = a/d
  #! si = b/d
  = (rotated_cs co si o scaled (V2 cfl cft)) spot p
//  = (rotated_cs co (si) o scaled (V2 cfl cft)) spot p

radial p = p - V2 0.2 (-0.3)
rotor p
  # (V2 x y) = radial p
  = V2 (~y) x
spotgrad p = lint RDarkRed RYellow (dfrspot radial p)
spotcurl p = lint RDarkRed RYellow (dfrspot rotor p)




/* 
rotatedAbout p0 ang  // of_a_texture
 = translated p0 o rotated_cs (cos ang) (sin ang) o translated (~p0)

*/















